from collections.abc import AsyncIterator, Sequence
from itertools import pairwise
from logging import getLogger
from typing import Any, Protocol, TypeVar, cast, final

from ..errors import WorkflowConstructionError
from ..packet import Packet
from ..processors.processor import Processor
from ..run_context import CtxT, RunContext
from ..typing.events import Event, ProcPacketOutEvent, ProcPayloadOutEvent
from ..typing.io import InT, OutT, ProcName
from ..utils.callbacks import is_method_overridden
from .workflow_processor import WorkflowProcessor

logger = getLogger(__name__)

_OutT_contra = TypeVar("_OutT_contra", contravariant=True)


class WorkflowLoopTerminator(Protocol[_OutT_contra, CtxT]):
    def __call__(
        self,
        out_packet: Packet[_OutT_contra],
        *,
        ctx: RunContext[CtxT],
        **kwargs: Any,
    ) -> bool: ...


class LoopedWorkflow(WorkflowProcessor[InT, OutT, CtxT]):
    def __init__(
        self,
        name: ProcName,
        subprocs: Sequence[Processor[Any, Any, CtxT]],
        exit_proc: Processor[Any, OutT, CtxT],
        recipients: list[ProcName] | None = None,
        max_iterations: int = 10,
    ) -> None:
        super().__init__(
            subprocs=subprocs,
            name=name,
            start_proc=subprocs[0],
            end_proc=exit_proc,
            recipients=recipients,
        )

        for prev_proc, proc in pairwise(subprocs):
            if prev_proc.out_type != proc.in_type:
                raise WorkflowConstructionError(
                    f"Output type {prev_proc.out_type} of subprocessor "
                    f"{prev_proc.name} does not match input type {proc.in_type} of "
                    f"subprocessor {proc.name}"
                )
        if subprocs[-1].out_type != subprocs[0].in_type:
            raise WorkflowConstructionError(
                "Looped workflow's last subprocessor output type "
                f"{subprocs[-1].out_type} does not match first subprocessor input "
                f"type {subprocs[0].in_type}"
            )

        self._max_iterations = max_iterations

    @property
    def max_iterations(self) -> int:
        return self._max_iterations

    def terminate_workflow_loop_impl(
        self, out_packet: Packet[OutT], *, ctx: RunContext[CtxT], **kwargs: Any
    ) -> bool:
        raise NotImplementedError

    def add_workflow_loop_terminator(
        self, func: WorkflowLoopTerminator[OutT, CtxT]
    ) -> WorkflowLoopTerminator[OutT, CtxT]:
        self.terminate_workflow_loop_impl = func

        return func

    @final
    def terminate_workflow_loop(
        self, out_packet: Packet[OutT], *, ctx: RunContext[CtxT], **kwargs: Any
    ) -> bool:
        base_cls = LoopedWorkflow[Any, Any, Any]
        if is_method_overridden("terminate_workflow_loop_impl", self, base_cls):
            return self.terminate_workflow_loop_impl(out_packet, ctx=ctx, **kwargs)

        return False

    async def _process(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> list[OutT]:
        packet = Packet(sender=self.name, payloads=in_args) if in_args else None

        for iteration_num in range(1, self._max_iterations + 1):
            for subproc in self.subprocs:
                logger.info(f"\n[Running subprocessor {subproc.name}]\n")

                packet = await subproc.run(
                    chat_inputs=chat_inputs,
                    in_packet=packet,
                    call_id=f"{call_id}/{subproc.name}/iter_{iteration_num}",
                    ctx=ctx,
                )

                logger.info(f"\n[Finished running subprocessor {subproc.name}]\n")

                if subproc is self.end_proc:
                    exit_packet = cast("Packet[OutT]", packet)
                    if self.terminate_workflow_loop(exit_packet, ctx=ctx):
                        return list(exit_packet.payloads)
                    if iteration_num == self._max_iterations:
                        logger.info(
                            f"Max iterations reached ({self._max_iterations}). "
                            "Exiting loop."
                        )
                        return list(exit_packet.payloads)

                chat_inputs = None

        raise RuntimeError("Looped workflow did not exit after max iterations.")

    @final
    async def _process_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> AsyncIterator[Event[Any]]:
        packet = Packet(sender=self.name, payloads=in_args) if in_args else None

        for iteration_num in range(1, self._max_iterations + 1):
            for subproc in self.subprocs:
                logger.info(f"\n[Running subprocessor {subproc.name}]\n")

                async for event in subproc.run_stream(
                    chat_inputs=chat_inputs,
                    in_packet=packet,
                    call_id=f"{call_id}/{subproc.name}/iter_{iteration_num}",
                    ctx=ctx,
                ):
                    yield event
                    if (
                        isinstance(event, ProcPacketOutEvent)
                        and event.src_name == subproc.name
                    ):
                        packet = event.data

                logger.info(f"\n[Finished running subprocessor {subproc.name}]\n")

                if subproc is self.end_proc:
                    exit_packet = cast("Packet[OutT]", packet)

                    if self.terminate_workflow_loop(exit_packet, ctx=ctx):
                        for p in exit_packet.payloads:
                            yield ProcPayloadOutEvent(
                                data=p, src_name=self.name, call_id=call_id
                            )
                        return

                    if iteration_num == self._max_iterations:
                        logger.info(
                            f"Max iterations reached ({self._max_iterations}). "
                            "Exiting loop."
                        )
                        for p in exit_packet.payloads:
                            yield ProcPayloadOutEvent(
                                data=p, src_name=self.name, call_id=call_id
                            )
                        return

                chat_inputs = None



import inspect
from collections.abc import Callable, Coroutine
from typing import Any

from .types import AsyncFunctionOrMethod, P, R


def is_bound_method(func: Callable[..., Any], self_candidate: Any) -> bool:
    return (inspect.ismethod(func) and (func.__self__ is self_candidate)) or hasattr(
        self_candidate, func.__name__
    )


def split_pos_args(
    call: AsyncFunctionOrMethod[P, R], args: tuple[Any, ...]
) -> tuple[Any | None, tuple[Any, ...]]:
    if not args:
        raise ValueError("No positional arguments passed.")
    maybe_self = args[0]
    if is_bound_method(call, maybe_self):
        # Case: Bound instance method with signature (self, inp, *rest)
        self_arg = args[0]
        remaining_args = args[1:]

        return self_arg, remaining_args

    # Case: Standalone function with signature (inp, *rest)
    if not args:
        raise ValueError(
            "Must pass an input (or a list of inputs) for a standalone function."
        )
    self_arg = None
    remaining_args = args

    return self_arg, remaining_args


def partial_callable(
    call: Callable[..., Coroutine[Any, Any, R]],
    self_obj: Any,
    *args: Any,
    **kwargs: Any,
) -> Callable[..., Coroutine[Any, Any, R]]:
    async def wrapper(inp: Any) -> R:
        if self_obj is not None:
            # `call` is a method
            return await call(self_obj, inp, *args, **kwargs)
        # `call` is a function
        return await call(inp, *args, **kwargs)

    return wrapper



import asyncio
import logging
from collections.abc import AsyncIterator
from types import TracebackType
from typing import Any, Protocol, TypeVar

from .typing.events import Event

logger = logging.getLogger(__name__)


_D_contra = TypeVar("_D_contra", contravariant=True)


class EventHandler(Protocol[_D_contra]):
    async def __call__(self, event: Event[_D_contra], **kwargs: Any) -> None: ...


class EventBus:
    def __init__(self) -> None:
        self._task_group: asyncio.TaskGroup | None = None

        self._routed_event_queues: dict[str, asyncio.Queue[Event[Any] | None]] = {}
        self._streamed_event_queue: asyncio.Queue[Event[Any] | None] = asyncio.Queue()

        self._event_handlers: dict[str, EventHandler[Any]] = {}
        self._handler_tasks: dict[str, asyncio.Task[None]] = {}

        self._final_result_fut: asyncio.Future[Any]

        self._stopping = False
        self._stopped_evt = asyncio.Event()

    def register_event_handler(self, dst_name: str, handler: EventHandler[Any]) -> None:
        if self._stopping:
            return

        # Prevent multiple concurrent handler tasks for the same destination
        if dst_name in self._handler_tasks and not self._handler_tasks[dst_name].done():
            raise RuntimeError(f"Handler already registered for {dst_name}")

        self._routed_event_queues.setdefault(dst_name, asyncio.Queue())

        self._event_handlers[dst_name] = handler
        if self._task_group is not None:
            self._handler_tasks[dst_name] = self._task_group.create_task(
                self._handle_events(dst_name), name=f"event-handler:{dst_name}"
            )

    async def post(self, event: Event[Any]) -> None:
        if self._stopping:
            return

        if event.dst_name is not None:
            queue = self._routed_event_queues[event.dst_name]
            await queue.put(event)

    async def push_to_stream(self, event: Event[Any]) -> None:
        if self._stopping:
            return

        await self._streamed_event_queue.put(event)

    async def stream_events(self) -> AsyncIterator[Event[Any]]:
        while True:
            event = await self._streamed_event_queue.get()
            if event is None:
                break
            yield event

    async def final_result(self) -> Any:
        return await self._final_result_fut

    async def __aenter__(self) -> "EventBus":
        self._task_group = asyncio.TaskGroup()
        await self._task_group.__aenter__()

        self._final_result_fut = asyncio.get_running_loop().create_future()

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> bool | None:
        await self.shutdown()

        ret: bool | None = False
        if self._task_group is not None:
            try:
                ret = await self._task_group.__aexit__(exc_type, exc, tb)
            finally:
                self._task_group = None

        # Fallback only: if finalize() already set the result/exception, don't override it.
        if not self._final_result_fut.done():
            if exc is not None:
                if isinstance(exc, asyncio.CancelledError):
                    self._final_result_fut.cancel()
                else:
                    self._final_result_fut.set_exception(exc)
            else:
                self._final_result_fut.cancel()

        return ret

    async def _handle_events(self, dst_name: str) -> None:
        handler = self._event_handlers[dst_name]

        while True:
            queue = self._routed_event_queues[dst_name]
            event = await queue.get()

            if event is None:
                break

            if self._final_result_fut.done():
                break

            try:
                await handler(event)
            except asyncio.CancelledError as err:
                # Cooperative cancellation: the whole TaskGroup is being cancelled)
                logger.info("Event handler cancelled for %s", dst_name)
                self.set_result(None, err=err)
                raise
            except Exception as err:
                # Unexpected error: only this handler is affected
                logger.exception("Error handling event for %s", dst_name)
                self.set_result(None, err)
                await self.shutdown()
                raise  # let TaskGroup propagate

    def set_result(
        self, result: Any, err: Exception | asyncio.CancelledError | None = None
    ) -> None:
        if not self._final_result_fut.done():
            if err and isinstance(err, asyncio.CancelledError):
                self._final_result_fut.cancel()
            elif err:
                self._final_result_fut.set_exception(err)
            else:
                self._final_result_fut.set_result(result)

    async def finalize(self, result: Any, err: Exception | None = None) -> None:
        self.set_result(result, err)
        await self.shutdown()

    async def shutdown(self) -> None:
        if self._stopping:
            await self._stopped_evt.wait()
            return
        self._stopping = True
        try:
            for queue in self._routed_event_queues.values():
                await queue.put(None)
            await self._streamed_event_queue.put(None)
        finally:
            self._stopped_evt.set()



import asyncio
import json
from collections.abc import AsyncIterator, Coroutine, Mapping, Sequence
from copy import deepcopy
from itertools import starmap
from logging import getLogger
from typing import Any, Generic, Protocol, TypedDict, cast, final

from pydantic import BaseModel

from grasp_agents.tracing_decorators import task

from .errors import AgentFinalAnswerError
from .llm import LLM
from .llm_agent_memory import LLMAgentMemory
from .run_context import CtxT, RunContext
from .typing.completion import Completion
from .typing.events import (
    Event,
    GenMessageEvent,
    ToolCallEvent,
    ToolMessageEvent,
    ToolOutputEvent,
    UserMessageEvent,
)
from .typing.message import AssistantMessage, ToolMessage, UserMessage
from .typing.tool import BaseTool, NamedToolChoice, ToolCall, ToolChoice
from .utils.callbacks import is_method_overridden
from .utils.streaming import EventStream, stream_concurrent

logger = getLogger(__name__)


class FinalAnswerChecker(Protocol[CtxT]):
    def __call__(
        self,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> str | None: ...


class BeforeGenerateHook(Protocol[CtxT]):
    async def __call__(
        self,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        extra_llm_settings: dict[str, Any],
    ) -> None: ...


class AfterGenerateHook(Protocol[CtxT]):
    async def __call__(
        self,
        gen_message: AssistantMessage,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
    ) -> None: ...


class ToolOutputConverter(Protocol[CtxT]):
    def __call__(
        self,
        tool_outputs: Sequence[Any],
        tool_calls: Sequence[ToolCall],
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> Sequence[ToolMessage | UserMessage]: ...


class CallArgs(TypedDict, total=False):
    ctx: RunContext[Any]
    call_id: str


class LLMPolicyExecutor(Generic[CtxT]):
    def __init__(
        self,
        *,
        agent_name: str,
        llm: LLM,
        memory: LLMAgentMemory,
        tools: list[BaseTool[BaseModel, Any, CtxT]] | None,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        max_turns: int,
        react_mode: bool = False,
        final_answer_type: type[BaseModel] = BaseModel,
        final_answer_as_tool_call: bool = False,
        stream_llm_responses: bool = True,
        stream_tools: bool = False,
    ) -> None:
        super().__init__()

        self._agent_name = agent_name
        self._max_turns = max_turns
        self._react_mode = react_mode

        self._llm = llm
        self._response_schema = response_schema
        self._response_schema_by_xml_tag = response_schema_by_xml_tag

        self.memory = memory

        self._final_answer_type = final_answer_type
        self._final_answer_as_tool_call = final_answer_as_tool_call
        self._final_answer_tool = self.get_final_answer_tool()

        self._stream_llm_responses = stream_llm_responses
        self._stream_tools = stream_tools

        tools_list: list[BaseTool[BaseModel, Any, CtxT]] | None = tools
        if tools and final_answer_as_tool_call:
            tools_list = tools + [self._final_answer_tool]
        self._tools = {t.name: t for t in tools_list} if tools_list else None

    @property
    def agent_name(self) -> str:
        return self._agent_name

    @property
    def max_turns(self) -> int:
        return self._max_turns

    @property
    def react_mode(self) -> bool:
        return self._react_mode

    @property
    def llm(self) -> LLM:
        return self._llm

    @property
    def response_schema(self) -> Any | None:
        return self._response_schema

    @response_schema.setter
    def response_schema(self, value: Any | None) -> None:
        self._response_schema = value

    @property
    def response_schema_by_xml_tag(self) -> Mapping[str, Any] | None:
        return self._response_schema_by_xml_tag

    @property
    def final_answer_as_tool_call(self) -> bool:
        return self._final_answer_as_tool_call

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, CtxT]]:
        return self._tools or {}

    def check_for_final_answer_impl(
        self,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> str | None:
        raise NotImplementedError

    @final
    def check_for_final_answer(
        self,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> str | None:
        if is_method_overridden("check_for_final_answer_impl", self):
            return self.check_for_final_answer_impl(ctx=ctx, call_id=call_id, **kwargs)

        if self._final_answer_as_tool_call:
            return self._get_final_answer_from_tool_call()

        return None

    async def on_before_generate_impl(
        self,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        extra_llm_settings: dict[str, Any],
    ) -> None:
        raise NotImplementedError

    @final
    async def on_before_generate(
        self,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        extra_llm_settings: dict[str, Any],
    ) -> None:
        if is_method_overridden("on_before_generate_impl", self):
            await self.on_before_generate_impl(
                ctx=ctx,
                call_id=call_id,
                num_turns=num_turns,
                extra_llm_settings=extra_llm_settings,
            )

    async def on_after_generate_impl(
        self,
        gen_message: AssistantMessage,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
    ) -> None:
        raise NotImplementedError

    @final
    async def on_after_generate(
        self,
        gen_message: AssistantMessage,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
    ) -> None:
        if is_method_overridden("on_after_generate_impl", self):
            await self.on_after_generate_impl(
                gen_message=gen_message,
                ctx=ctx,
                call_id=call_id,
                num_turns=num_turns,
            )

    @task(name="generate")  # type: ignore
    async def generate_message_stream(
        self,
        *,
        tool_choice: ToolChoice | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
        extra_llm_settings: dict[str, Any],
    ) -> AsyncIterator[Event[Any]]:
        completion: Completion | None = None

        if self._stream_llm_responses:
            llm_stream = self.llm.generate_completion_stream(
                self.memory.messages,
                response_schema=self.response_schema,
                response_schema_by_xml_tag=self.response_schema_by_xml_tag,
                tools=self.tools,
                tool_choice=tool_choice,
                proc_name=self.agent_name,
                call_id=call_id,
                **extra_llm_settings,
            )
            llm_stream_post = self.llm.postprocess_event_stream(llm_stream)
            llm_stream_wrapped = EventStream[Completion](llm_stream_post, Completion)
            async for event in llm_stream_wrapped:
                yield event
            completion = await llm_stream_wrapped.final_data()

        else:
            completion = await self.llm.generate_completion(
                self.memory.messages,
                response_schema=self.response_schema,
                response_schema_by_xml_tag=self.response_schema_by_xml_tag,
                tools=self.tools,
                tool_choice=tool_choice,
                proc_name=self.agent_name,
                call_id=call_id,
                **extra_llm_settings,
            )

        yield GenMessageEvent(
            src_name=self.agent_name, call_id=call_id, data=completion.message
        )
        self.memory.update([completion.message])
        self._process_completion(completion, ctx=ctx, call_id=call_id)

    async def generate_message(
        self,
        *,
        tool_choice: ToolChoice | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
        extra_llm_settings: dict[str, Any],
    ) -> AssistantMessage:
        gen_message: AssistantMessage | None = None
        async for event in self.generate_message_stream(
            tool_choice=tool_choice,
            extra_llm_settings=extra_llm_settings,
            ctx=ctx,
            call_id=call_id,
        ):
            if isinstance(event, GenMessageEvent):
                gen_message = event.data

        return cast("AssistantMessage", gen_message)

    def tool_outputs_to_messages_impl(
        self,
        tool_outputs: Sequence[Any],
        tool_calls: Sequence[ToolCall],
        *,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> Sequence[ToolMessage | UserMessage]:
        raise NotImplementedError

    def tool_outputs_to_messages_default(
        self, tool_outputs: Sequence[Any], tool_calls: Sequence[ToolCall]
    ) -> Sequence[ToolMessage | UserMessage]:
        return list(
            starmap(
                ToolMessage.from_tool_output, zip(tool_outputs, tool_calls, strict=True)
            )
        )

    @final
    def tool_outputs_to_messages(
        self,
        tool_outputs: Sequence[Any],
        tool_calls: Sequence[ToolCall],
        *,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> Sequence[ToolMessage | UserMessage]:
        if is_method_overridden("tool_outputs_to_messages_impl", self):
            return self.tool_outputs_to_messages_impl(
                tool_outputs, tool_calls, ctx=ctx, call_id=call_id
            )
        return self.tool_outputs_to_messages_default(tool_outputs, tool_calls)

    async def _get_tool_outputs(
        self, calls: Sequence[ToolCall], ctx: RunContext[CtxT], call_id: str
    ) -> Sequence[Any]:
        corouts: list[Coroutine[Any, Any, BaseModel]] = []
        for call in calls:
            tool = self.tools[call.tool_name]
            args = json.loads(call.tool_arguments)
            corouts.append(tool(ctx=ctx, call_id=call_id, **args))

        return await asyncio.gather(*corouts)

    async def call_tools_stream(
        self, calls: Sequence[ToolCall], ctx: RunContext[CtxT], call_id: str
    ) -> AsyncIterator[Event[Any]]:
        for call in calls:
            yield ToolCallEvent(src_name=self.agent_name, call_id=call_id, data=call)

        if self._stream_tools:
            streams: list[AsyncIterator[Event[Any]]] = []
            for call in calls:
                tool = self.tools[call.tool_name]
                args = json.loads(call.tool_arguments)
                streams.append(
                    tool.run_stream(inp=tool.in_type(**args), ctx=ctx, call_id=call_id)
                )

            outputs_map: dict[int, Any] = {}
            async for idx, event in stream_concurrent(streams):
                if isinstance(event, ToolOutputEvent):
                    outputs_map[idx] = event.data
                else:
                    yield event
            outputs = [outputs_map[idx] for idx in sorted(outputs_map)]

        else:
            outputs = await self._get_tool_outputs(calls, ctx=ctx, call_id=call_id)

        tool_messages = self.tool_outputs_to_messages(
            outputs, calls, ctx=ctx, call_id=call_id
        )

        for tool_message, call in zip(tool_messages, calls, strict=True):
            if isinstance(tool_message, UserMessage):
                yield UserMessageEvent(
                    src_name=call.tool_name, call_id=call_id, data=tool_message
                )
            else:
                yield ToolMessageEvent(
                    src_name=call.tool_name, call_id=call_id, data=tool_message
                )

        self.memory.update(tool_messages)

        if ctx.printer:
            ctx.printer.print_messages(
                tool_messages, agent_name=self.agent_name, call_id=call_id
            )

    async def call_tools(
        self, calls: Sequence[ToolCall], ctx: RunContext[CtxT], call_id: str
    ) -> Sequence[ToolMessage | UserMessage]:
        tool_messages: list[ToolMessage | UserMessage] = []
        async for event in self.call_tools_stream(calls, ctx=ctx, call_id=call_id):
            if isinstance(event, ToolMessageEvent | UserMessageEvent):
                tool_messages.append(event.data)

        return tool_messages

    def _get_final_answer_from_tool_call(self) -> str | None:
        msgs = self.memory.messages
        if (
            msgs
            and isinstance(msgs[-1], AssistantMessage)
            and msgs[-1].tool_calls
            and msgs[-1].tool_calls[0].tool_name == self._final_answer_tool.name
        ):
            return msgs[-1].tool_calls[0].tool_arguments
        return None

    def _get_final_answer_from_message(self) -> str | None:
        msgs = self.memory.messages
        if msgs and isinstance(msgs[-1], AssistantMessage) and msgs[-1].content:
            return msgs[-1].content
        return None

    def get_final_answer(self) -> str | None:
        if self._final_answer_as_tool_call:
            return self._get_final_answer_from_tool_call()
        return self._get_final_answer_from_message()

    @task(name="force_generate_final_answer")  # type: ignore
    async def _force_generate_final_answer_stream(
        self, ctx: RunContext[CtxT], call_id: str, extra_llm_settings: dict[str, Any]
    ) -> AsyncIterator[Event[Any]]:
        # NOTE: Might not need the user message when forcing the tool call
        user_message = UserMessage.from_text(
            "Exceeded the maximum number of turns: provide a final answer now!",
        )
        self.memory.update([user_message])
        yield UserMessageEvent(
            src_name=self.agent_name, call_id=call_id, data=user_message
        )
        if ctx.printer:
            ctx.printer.print_messages(
                [user_message], agent_name=self.agent_name, call_id=call_id
            )

        tool_choice = (
            NamedToolChoice(name=self._final_answer_tool.name)
            if self._final_answer_as_tool_call
            else None
        )
        async for event in self.generate_message_stream(
            tool_choice=tool_choice,
            ctx=ctx,
            call_id=call_id,
            extra_llm_settings=extra_llm_settings,
        ):
            yield event

        final_answer = self.get_final_answer()
        if final_answer is None:
            raise AgentFinalAnswerError(proc_name=self.agent_name, call_id=call_id)

    async def _force_generate_final_answer(
        self, ctx: RunContext[CtxT], call_id: str, extra_llm_settings: dict[str, Any]
    ) -> str:
        async for _ in self._force_generate_final_answer_stream(
            ctx=ctx, call_id=call_id, extra_llm_settings=extra_llm_settings
        ):
            pass
        return cast("str", self.get_final_answer())

    async def execute_stream(
        self,
        ctx: RunContext[CtxT],
        call_id: str,
        extra_llm_settings: dict[str, Any] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        call_kwargs: CallArgs = CallArgs(ctx=ctx, call_id=call_id)

        turns = 0
        gen_message: AssistantMessage | None = None

        # 1. Generate the first message and update memory

        _extra_llm_settings = deepcopy(extra_llm_settings or {})
        await self.on_before_generate(
            extra_llm_settings=_extra_llm_settings, num_turns=turns, **call_kwargs
        )

        tool_choice: ToolChoice | None = None
        if self.tools:
            tool_choice = "none" if self.react_mode else "auto"
        tool_choice = _extra_llm_settings.pop("tool_choice", tool_choice)

        async for event in self.generate_message_stream(
            tool_choice=tool_choice,
            extra_llm_settings=_extra_llm_settings,
            **call_kwargs,
        ):
            if isinstance(event, GenMessageEvent):
                gen_message = event.data
            yield event
        gen_message = cast("AssistantMessage", gen_message)

        await self.on_after_generate(gen_message, num_turns=turns, **call_kwargs)

        if not self.tools:
            # No tools to call, return the content of the generated message
            return

        while True:
            # 2. Check if we have a final answer

            final_answer = self.check_for_final_answer(
                ctx=ctx, call_id=call_id, num_turns=turns
            )
            if final_answer is not None:
                return

            if turns >= self.max_turns:
                async for event in self._force_generate_final_answer_stream(
                    extra_llm_settings=_extra_llm_settings, **call_kwargs
                ):
                    yield event
                logger.info(
                    f"Max turns reached: {self.max_turns}. Exiting the tool call loop."
                )
                return

            # 3. Call tools and update memory

            if gen_message.tool_calls:
                async for event in self.call_tools_stream(
                    gen_message.tool_calls, **call_kwargs
                ):
                    yield event

            # 4. Generate the next message and update memory

            _extra_llm_settings = deepcopy(extra_llm_settings or {})
            await self.on_before_generate(
                extra_llm_settings=_extra_llm_settings, num_turns=turns, **call_kwargs
            )

            if self.react_mode and gen_message.tool_calls:
                tool_choice = "none"
            elif self.react_mode:
                tool_choice = "required"
            else:
                tool_choice = "auto"
            tool_choice = _extra_llm_settings.pop("tool_choice", tool_choice)

            async for event in self.generate_message_stream(
                tool_choice=tool_choice,
                extra_llm_settings=_extra_llm_settings,
                **call_kwargs,
            ):
                if isinstance(event, GenMessageEvent):
                    gen_message = event.data
                yield event

            await self.on_after_generate(gen_message, num_turns=turns, **call_kwargs)

            turns += 1

    async def execute(
        self,
        ctx: RunContext[CtxT],
        call_id: str,
        extra_llm_settings: dict[str, Any] | None = None,
    ) -> str:
        async for _ in self.execute_stream(
            ctx=ctx, call_id=call_id, extra_llm_settings=extra_llm_settings
        ):
            pass
        return cast("str", self.get_final_answer())

    def get_final_answer_tool(self) -> BaseTool[BaseModel, None, Any]:
        class FinalAnswerTool(BaseTool[self._final_answer_type, None, Any]):
            name: str = "final_answer"
            description: str = (
                "You must call this tool to provide the final answer. "
                "DO NOT output your answer before calling the tool. "
            )

            async def run(
                self,
                inp: BaseModel,
                *,
                ctx: RunContext[Any] | None = None,
                call_id: str | None = None,
            ) -> None:
                return None

        return FinalAnswerTool()

    def _process_completion(
        self, completion: Completion, *, ctx: RunContext[CtxT], call_id: str
    ) -> None:
        ctx.completions[self.agent_name].append(completion)
        ctx.usage_tracker.update(
            agent_name=self.agent_name,
            completions=[completion],
            model_name=self.llm.model_name,
        )
        if ctx.printer:
            ctx.printer.print_messages(
                [completion.message],
                usages=[completion.usage],
                agent_name=self.agent_name,
                call_id=call_id,
            )

    # def wrapped_generate_message_stream(
    #     self,
    #     *,
    #     tool_choice: ToolChoice | None = None,
    #     ctx: RunContext[CtxT],
    #     call_id: str,
    #     extra_llm_settings: dict[str, Any],
    # ) -> EventStream[AssistantMessage]:
    #     stream = self.generate_message_stream(
    #         tool_choice=tool_choice,
    #         ctx=ctx,
    #         call_id=call_id,
    #         extra_llm_settings=extra_llm_settings,
    #     )
    #     return EventStream[AssistantMessage](stream, AssistantMessage)



from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel

from .completion import Completion, Usage
from .completion_chunk import CompletionChunk
from .content import Content
from .message import AssistantMessage, Message, SystemMessage, ToolMessage, UserMessage
from .tool import BaseTool, ToolChoice


class Converters(ABC):
    @staticmethod
    @abstractmethod
    def to_system_message(system_message: SystemMessage, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_system_message(raw_message: Any, **kwargs: Any) -> SystemMessage:
        pass

    @staticmethod
    @abstractmethod
    def to_user_message(user_message: UserMessage, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_user_message(raw_message: Any, **kwargs: Any) -> UserMessage:
        pass

    @staticmethod
    @abstractmethod
    def to_assistant_message(assistant_message: AssistantMessage, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_completion_usage(raw_usage: Any, **kwargs: Any) -> Usage:
        pass

    @staticmethod
    @abstractmethod
    def from_assistant_message(raw_message: Any, **kwargs: Any) -> AssistantMessage:
        pass

    @staticmethod
    @abstractmethod
    def to_tool_message(tool_message: ToolMessage, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_tool_message(raw_message: Any, **kwargs: Any) -> ToolMessage:
        pass

    @classmethod
    def to_message(cls, message: Message, **kwargs: Any) -> Any:
        if isinstance(message, UserMessage):
            return cls.to_user_message(message, **kwargs)
        if isinstance(message, AssistantMessage):
            return cls.to_assistant_message(message, **kwargs)
        if isinstance(message, ToolMessage):
            return cls.to_tool_message(message, **kwargs)

        return cls.to_system_message(message, **kwargs)

    @staticmethod
    @abstractmethod
    def to_tool(
        tool: BaseTool[BaseModel, Any, Any], strict: bool | None = None, **kwargs: Any
    ) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def to_tool_choice(tool_choice: ToolChoice, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def to_content(content: Content, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_content(raw_content: Any, **kwargs: Any) -> Content:
        pass

    @staticmethod
    @abstractmethod
    def to_completion(completion: Completion, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_completion(raw_completion: Any, **kwargs: Any) -> Completion:
        pass

    @staticmethod
    @abstractmethod
    def to_completion_chunk(chunk: CompletionChunk, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_completion_chunk(raw_chunk: Any, **kwargs: Any) -> CompletionChunk:
        pass



import logging
from collections import defaultdict
from collections.abc import AsyncIterator, Mapping
from dataclasses import dataclass, field
from typing import Any, ClassVar, cast

import litellm
from litellm.litellm_core_utils.get_supported_openai_params import (
    get_supported_openai_params,  # type: ignore[no-redef]
)
from litellm.litellm_core_utils.streaming_handler import CustomStreamWrapper
from litellm.router import Router
from litellm.types.llms.anthropic import AnthropicThinkingParam
from litellm.utils import (
    supports_parallel_function_calling,
    supports_prompt_caching,
    supports_reasoning,
    supports_response_schema,
    supports_tool_choice,
)

# from openai.lib.streaming.chat import ChunkEvent as OpenAIChunkEvent
from pydantic import BaseModel

from ..cloud_llm import APIProvider, CloudLLM
from ..openai.openai_llm import OpenAILLMSettings
from ..typing.tool import BaseTool
from . import (
    LiteLLMCompletion,
    LiteLLMCompletionChunk,
    OpenAIMessageParam,
    OpenAIToolChoiceOptionParam,
    OpenAIToolParam,
)
from .converters import LiteLLMConverters

logger = logging.getLogger(__name__)


class LiteLLMSettings(OpenAILLMSettings, total=False):
    thinking: AnthropicThinkingParam | None


LiteLLMModelName = str


@dataclass(frozen=True)
class LiteLLM(CloudLLM):
    llm_settings: LiteLLMSettings | None = None
    converters: ClassVar[LiteLLMConverters] = LiteLLMConverters()

    client_timeout: float = 60.0
    max_client_retries: int = 2

    # Drop unsupported OpenAI params
    drop_params: bool = False
    additional_drop_params: list[str] | None = None
    allowed_openai_params: list[str] | None = None
    # Mock LLM response for testing
    mock_response: str | None = None
    # Fallback models to use if the main model fails
    fallbacks: list[LiteLLMModelName] = field(default_factory=list[LiteLLMModelName])
    # Mock falling back to other models in the fallbacks list for testing
    mock_testing_fallbacks: bool = False

    router: Router = field(init=False)

    _lite_llm_completion_params: dict[str, Any] = field(
        default_factory=dict[str, Any], init=False, repr=False, compare=False
    )

    def __post_init__(self) -> None:
        super().__post_init__()

        self._lite_llm_completion_params.update(
            {
                "drop_params": self.drop_params,
                "additional_drop_params": self.additional_drop_params,
                "allowed_openai_params": self.allowed_openai_params,
                "mock_response": self.mock_response,
                "mock_testing_fallbacks": self.mock_testing_fallbacks,
                # "max_retries": self.max_client_retries,
                # "timeout": self.client_timeout,
                # "deployment_id": deployment_id,
                # "api_version": api_version,
            }
        )

        _api_provider = self.api_provider
        try:
            _, provider_name, api_key, api_base = litellm.get_llm_provider(  # type: ignore[no-untyped-call]
                self.model_name
            )
            _api_provider = APIProvider(
                name=provider_name, api_key=api_key, base_url=api_base
            )
        except Exception as exc:
            if self.api_provider is not None:
                self._lite_llm_completion_params["api_key"] = self.api_provider.get(
                    "api_key"
                )
                self._lite_llm_completion_params["api_base"] = self.api_provider.get(
                    "api_base"
                )
            else:
                raise ValueError(
                    f"Failed to retrieve a LiteLLM supported API provider for model "
                    f"'{self.model_name}' and no custom API provider was specified."
                ) from exc

        if (
            self.apply_response_schema_via_provider
            and not self.supports_response_schema
        ):
            raise ValueError(
                f"Model '{self.model_name}' does not support response schema "
                "natively. Please set `apply_response_schema_via_provider=False`"
            )

        if self.http_client is not None:
            raise NotImplementedError(
                "Custom HTTP clients are not yet supported when using LiteLLM."
            )

        main_litellm_model = {
            "model_name": self.model_name,
            "litellm_params": {"model": self.model_name},
        }
        fallback_litellm_models = [
            {"model_name": fb, "litellm_params": {"model": fb}} for fb in self.fallbacks
        ]

        _router = Router(
            model_list=[main_litellm_model, *fallback_litellm_models],
            fallbacks=[{self.model_name: self.fallbacks}],
            num_retries=self.max_client_retries,
            timeout=self.client_timeout,
        )

        object.__setattr__(self, "router", _router)
        object.__setattr__(self, "api_provider", _api_provider)

    def get_supported_openai_params(self) -> list[Any] | None:
        return get_supported_openai_params(  # type: ignore[no-untyped-call]
            model=self.model_name, request_type="chat_completion"
        )

    @property
    def supports_reasoning(self) -> bool:
        return supports_reasoning(model=self.model_name)

    @property
    def supports_parallel_function_calling(self) -> bool:
        return supports_parallel_function_calling(model=self.model_name)

    @property
    def supports_prompt_caching(self) -> bool:
        return supports_prompt_caching(model=self.model_name)

    @property
    def supports_response_schema(self) -> bool:
        return supports_response_schema(model=self.model_name)

    @property
    def supports_tool_choice(self) -> bool:
        return supports_tool_choice(model=self.model_name)

    async def _get_api_completion(
        self,
        api_messages: list[OpenAIMessageParam],
        api_tools: list[OpenAIToolParam] | None = None,
        api_tool_choice: OpenAIToolChoiceOptionParam | None = None,
        api_response_schema: type | None = None,
        **api_llm_settings: Any,
    ) -> LiteLLMCompletion:
        completion = await self.router.acompletion(  # type: ignore[no-untyped-call]
            model=self.model_name,
            messages=api_messages,  # type: ignore[arg-type]
            tools=api_tools,
            tool_choice=api_tool_choice,  # type: ignore[arg-type]
            response_format=api_response_schema,
            stream=False,
            **self._lite_llm_completion_params,
            **api_llm_settings,
        )
        completion = cast("LiteLLMCompletion", completion)

        # Should not be needed in litellm>=1.74
        completion._hidden_params["response_cost"] = litellm.completion_cost(completion)  # type: ignore[no-untyped-call]

        return completion

    async def _get_api_completion_stream(
        self,
        api_messages: list[OpenAIMessageParam],
        api_tools: list[OpenAIToolParam] | None = None,
        api_tool_choice: OpenAIToolChoiceOptionParam | None = None,
        api_response_schema: type | None = None,
        **api_llm_settings: Any,
    ) -> AsyncIterator[LiteLLMCompletionChunk]:
        # Ensure usage is included in the streamed responses
        stream_options = dict(api_llm_settings.get("stream_options") or {})
        stream_options["include_usage"] = True
        _api_llm_settings = api_llm_settings | {"stream_options": stream_options}

        stream = await self.router.acompletion(  # type: ignore[no-untyped-call]
            model=self.model_name,
            messages=api_messages,  # type: ignore[arg-type]
            tools=api_tools,
            tool_choice=api_tool_choice,  # type: ignore[arg-type]
            response_format=api_response_schema,
            stream=True,
            **self._lite_llm_completion_params,
            **_api_llm_settings,
        )
        stream = cast("CustomStreamWrapper", stream)

        tc_indices: dict[int, set[int]] = defaultdict(set)

        # Need to wrap the iterator to make it work with decorators
        async def iterator() -> AsyncIterator[LiteLLMCompletionChunk]:
            async for completion_chunk in stream:
                # Fix tool call indices to be unique within each choice
                if completion_chunk is not None:
                    for n, choice in enumerate(completion_chunk.choices):
                        for tc in choice.delta.tool_calls or []:
                            # Tool call ID is not None only when it is a new tool call
                            if tc.id and tc.index in tc_indices[n]:
                                tc.index = max(tc_indices[n]) + 1
                            tc_indices[n].add(tc.index)

                    yield completion_chunk

        return iterator()

    def combine_completion_chunks(
        self,
        completion_chunks: list[LiteLLMCompletionChunk],
        response_schema: Any | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
    ) -> LiteLLMCompletion:
        combined_chunk = cast(
            "LiteLLMCompletion",
            litellm.stream_chunk_builder(completion_chunks),  # type: ignore[no-untyped-call]
        )
        # Should not be needed in litellm>=1.74
        combined_chunk._hidden_params["response_cost"] = litellm.completion_cost(  # type: ignore[no-untyped-call]
            combined_chunk
        )

        return combined_chunk



from logging import getLogger
from typing import Any

logger = getLogger(__name__)


def is_method_overridden(
    method_name: str, self: object, base_cls: type[Any] | None = None
) -> bool:
    """
    Check if a method is overridden in a subclass compared to a base class
    or if it is defined directly in the instance's __dict__ (e.g. via a decorator).
    """
    child_cls = type(self)
    if not hasattr(child_cls, method_name):
        raise AttributeError(f"{child_cls} has no method named {method_name}")

    set_via_decorator = method_name in self.__dict__

    overriden_in_child = False
    if base_cls is not None:
        overriden_in_child = hasattr(base_cls, method_name) and getattr(
            child_cls, method_name
        ) is not getattr(base_cls, method_name)

    return set_via_decorator or overriden_in_child



from typing import cast

from ..errors import CompletionError
from ..typing.completion_chunk import (
    CompletionChunk,
    CompletionChunkDelta,
    CompletionChunkDeltaToolCall,
)
from ..typing.message import Role
from . import OpenAICompletionChunk
from .completion_converters import from_api_completion_usage


def from_api_completion_chunk(
    api_completion_chunk: OpenAICompletionChunk, name: str | None = None
) -> CompletionChunk:
    if api_completion_chunk.choices is None:  # type: ignore
        raise CompletionError(
            f"Completion chunk API error: "
            f"{getattr(api_completion_chunk, 'error', None)}"
        )

    if len(api_completion_chunk.choices) > 1:
        raise CompletionError("Multiple choices are not supported")
    api_choice = api_completion_chunk.choices[0]

    # Some providers return None for the message when finish_reason is other than "stop"
    finish_reason = api_choice.finish_reason
    if api_choice.delta is None:  # type: ignore
        raise CompletionError(
            "API returned None for delta content in completion chunk "
            f"with finish_reason: {finish_reason}."
        )

    # if api_choice.delta.content is None:
    #     raise CompletionError(
    #         "API returned None for delta content in completion chunk "
    #         f"with finish_reason: {finish_reason}."
    #     )

    delta = CompletionChunkDelta(
        content=api_choice.delta.content,
        refusal=api_choice.delta.refusal,
        role=cast("Role", api_choice.delta.role),
        tool_calls=[
            CompletionChunkDeltaToolCall(
                id=tool_call.id,
                index=tool_call.index,
                tool_name=tool_call.function.name,
                tool_arguments=tool_call.function.arguments,
            )
            for tool_call in (api_choice.delta.tool_calls or [])
            if tool_call.function
        ],
    )

    usage = (
        from_api_completion_usage(api_completion_chunk.usage)
        if api_completion_chunk.usage
        else None
    )

    return CompletionChunk(
        id=api_completion_chunk.id,
        model=api_completion_chunk.model,
        name=name,
        created=api_completion_chunk.created,
        system_fingerprint=api_completion_chunk.system_fingerprint,
        delta=delta,
        finish_reason=finish_reason,
        logprobs=api_choice.logprobs,
        usage=usage,
    )



import logging
from collections.abc import AsyncIterator, Sequence
from itertools import pairwise
from typing import Any, cast

from ..errors import WorkflowConstructionError
from ..packet import Packet
from ..processors.processor import Processor
from ..run_context import CtxT, RunContext
from ..typing.events import Event, ProcPacketOutEvent, ProcPayloadOutEvent
from ..typing.io import InT, OutT, ProcName
from .workflow_processor import WorkflowProcessor

logger = logging.getLogger(__name__)


class SequentialWorkflow(WorkflowProcessor[InT, OutT, CtxT]):
    def __init__(
        self,
        name: ProcName,
        subprocs: Sequence[Processor[Any, Any, CtxT]],
        recipients: list[ProcName] | None = None,
    ) -> None:
        super().__init__(
            subprocs=subprocs,
            start_proc=subprocs[0],
            end_proc=subprocs[-1],
            name=name,
            recipients=recipients,
        )

        for prev_proc, proc in pairwise(subprocs):
            if prev_proc.out_type != proc.in_type:
                raise WorkflowConstructionError(
                    f"Output type {prev_proc.out_type} of subprocessor {prev_proc.name}"
                    f" does not match input type {proc.in_type} of subprocessor"
                    f" {proc.name}"
                )

    async def _process(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> list[OutT]:
        packet = Packet(sender=self.name, payloads=in_args) if in_args else None

        for subproc in self.subprocs:
            logger.info(f"\n[Running subprocessor {subproc.name}]\n")

            packet = await subproc.run(
                chat_inputs=chat_inputs,
                in_packet=packet,
                call_id=f"{call_id}/{subproc.name}",
                ctx=ctx,
            )
            chat_inputs = None

            logger.info(f"\n[Finished running subprocessor {subproc.name}]\n")

        return list(cast("Packet[OutT]", packet).payloads)

    async def _process_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> AsyncIterator[Event[Any]]:
        packet = Packet(sender=self.name, payloads=in_args) if in_args else None

        for subproc in self.subprocs:
            logger.info(f"\n[Running subprocessor {subproc.name}]\n")

            async for event in subproc.run_stream(
                chat_inputs=chat_inputs,
                in_packet=packet,
                call_id=f"{call_id}/{subproc.name}",
                ctx=ctx,
            ):
                yield event
                if (
                    isinstance(event, ProcPacketOutEvent)
                    and event.src_name == subproc.name
                ):
                    packet = event.data

            if subproc is self.end_proc:
                out_packet = cast("Packet[OutT]", packet)
                for p in out_packet.payloads:
                    yield ProcPayloadOutEvent(
                        data=p, src_name=self.name, call_id=call_id
                    )

            chat_inputs = None

            logger.info(f"\n[Finished running subprocessor {subproc.name}]\n")



import os
from logging import getLogger

from openinference.instrumentation.litellm import (
    LiteLLMInstrumentor as OILiteLLMInstrumentor,
)
from openinference.instrumentation.openai import (
    OpenAIInstrumentor as OIOpenAIInstrumentor,
)
from openinference.instrumentation.openllmetry import OpenInferenceSpanProcessor
from opentelemetry import trace as trace_api
from opentelemetry.sdk.trace import TracerProvider

from phoenix.otel import BatchSpanProcessor, HTTPSpanExporter, SimpleSpanProcessor

from .exporters import LLM_PROVIDER_NAMES, FilteringExporter

logger = getLogger(__name__)


def init_phoenix(
    batch: bool = False,
    use_litellm_instr: bool = True,
    use_llm_provider_instr: bool = True,
):
    collector_endpoint = os.getenv("TELEMETRY_COLLECTOR_HTTP_ENDPOINT")

    if not collector_endpoint:
        logger.warning(
            "TELEMETRY_COLLECTOR_HTTP_ENDPOINT not set, cannot initialize Phoenix"
        )
        return

    # 1) Get or create tracer provider

    tracer_provider = trace_api.get_tracer_provider()

    if not isinstance(tracer_provider, TracerProvider):
        logger.info("No existing tracer provider, cannot initialize Phoenix")
        return

    # 2) Convert spans to OpenInference format expected by Phoenix
    tracer_provider.add_span_processor(OpenInferenceSpanProcessor())

    # 3) Export to Phoenix backend
    # Use FilteringExporter to block LLM provider spans that are
    # already captured by OpenInference instrumentations
    blocklist: set[str] = (
        LLM_PROVIDER_NAMES if use_llm_provider_instr or use_litellm_instr else set()
    )
    exporter = FilteringExporter(
        inner=HTTPSpanExporter(endpoint=collector_endpoint, headers=None),
        blocklist=blocklist,
    )
    if batch:
        span_processor = BatchSpanProcessor(span_exporter=exporter)
    else:
        span_processor = SimpleSpanProcessor(span_exporter=exporter)
    tracer_provider.add_span_processor(span_processor)

    # 4) Replace OpenTelemetry instrumentations with OpenInference ones for better
    # compatibility with the Phoenix backend

    if use_litellm_instr:
        OILiteLLMInstrumentor().instrument(tracer_provider=tracer_provider)
    if use_llm_provider_instr:
        OIOpenAIInstrumentor().instrument(tracer_provider=tracer_provider)



import logging
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, AsyncIterator, Callable, Sequence
from copy import deepcopy
from functools import wraps
from typing import Any, ClassVar, Generic, Self, TypeVar, cast, final
from uuid import uuid4

from pydantic import BaseModel

from ..errors import ProcRunError
from ..generics_utils import AutoInstanceAttributesMixin
from ..memory import DummyMemory, Memory
from ..packet import Packet
from ..run_context import CtxT, RunContext
from ..typing.events import (
    DummyEvent,
    Event,
    ProcPacketOutEvent,
    ProcStreamingErrorData,
    ProcStreamingErrorEvent,
    ToolOutputEvent,
)
from ..typing.io import InT, OutT, ProcName
from ..typing.tool import BaseTool

logger = logging.getLogger(__name__)


F = TypeVar("F", bound=Callable[..., AsyncIterator[Event[Any]]])


def with_retry(func: F) -> F:
    @wraps(func)
    async def wrapper(
        self: "BaseProcessor[Any, Any, Any]", *args: Any, **kwargs: Any
    ) -> AsyncIterator[Event[Any]]:
        call_id = kwargs.get("call_id", "unknown")

        n_attempt = 0
        while n_attempt <= self.max_retries:
            try:
                async for event in func(self, *args, **kwargs):
                    yield event
                return

            except Exception as err:
                n_attempt += 1

                err_data = ProcStreamingErrorData(error=err, call_id=call_id)
                yield ProcStreamingErrorEvent(
                    data=err_data, src_name=self.name, call_id=call_id
                )
                err_message = (
                    f"Processor run failed [proc_name={self.name}; call_id={call_id}]"
                )
                if n_attempt > self.max_retries:
                    raise ProcRunError(proc_name=self.name, call_id=call_id) from err

                logger.warning(f"{err_message} (retry attempt {n_attempt}):\n{err}")

    return cast("F", wrapper)


class BaseProcessor(AutoInstanceAttributesMixin, ABC, Generic[InT, OutT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        name: ProcName,
        max_retries: int = 0,
        memory: Memory | None = None,
        recipients: Sequence[ProcName] | None = None,
        **kwargs: Any,
    ) -> None:
        self._in_type: type[InT]
        self._out_type: type[OutT]

        super().__init__()

        self._name = name
        self._max_retries = max_retries
        self._memory: Memory = memory or DummyMemory()
        self.recipients = recipients

    @property
    def in_type(self) -> type[InT]:
        return self._in_type

    @property
    def out_type(self) -> type[OutT]:
        return self._out_type

    @property
    def name(self) -> ProcName:
        return self._name

    @property
    def memory(self) -> Memory:
        return self._memory

    @property
    def max_retries(self) -> int:
        return self._max_retries

    def generate_call_id(self, call_id: str | None) -> str:
        if call_id is None:
            return str(uuid4())[:6] + "_" + self.name
        return call_id

    def copy(self) -> Self:
        return deepcopy(self)

    @abstractmethod
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT]:
        pass

    @abstractmethod
    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncGenerator[Event[Any], None]:
        yield DummyEvent()

    @final
    def as_tool(
        self, tool_name: str, tool_description: str, reset_memory_on_run: bool = True
    ) -> BaseTool[InT, OutT, Any]:  # type: ignore[override]
        # TODO: stream tools
        processor_instance = self
        in_type = processor_instance.in_type
        out_type = processor_instance.out_type
        if not issubclass(in_type, BaseModel):
            raise TypeError(
                "Cannot create a tool from an agent with "
                f"non-BaseModel input type: {in_type}"
            )

        class ProcessorTool(BaseTool[in_type, out_type, Any]):
            name: str = tool_name
            description: str = tool_description

            async def run(
                self,
                inp: InT,
                *,
                call_id: str | None = None,
                ctx: RunContext[CtxT] | None = None,
            ) -> OutT:
                if reset_memory_on_run:
                    processor_instance.memory.reset()

                result = await processor_instance.run(
                    in_args=inp, call_id=call_id, ctx=ctx
                )

                return result.payloads[0]

            async def run_stream(
                self,
                inp: InT,
                *,
                call_id: str | None = None,
                ctx: RunContext[CtxT] | None = None,
            ) -> AsyncIterator[Event[Any]]:
                if reset_memory_on_run:
                    processor_instance.memory.reset()

                async for event in processor_instance.run_stream(
                    in_args=inp, call_id=call_id, ctx=ctx
                ):
                    if (
                        isinstance(event, ProcPacketOutEvent)
                        and event.src_name == processor_instance.name
                    ):
                        yield ToolOutputEvent(
                            data=event.data.payloads[0],
                            src_name=processor_instance.name,
                            call_id=call_id,
                        )
                    else:
                        yield event

        return ProcessorTool()



import fnmatch
import logging
import os
from collections.abc import AsyncIterator, Iterable, Mapping
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, ClassVar, Literal

from openai import AsyncOpenAI, AsyncStream
from openai._types import NOT_GIVEN  # type: ignore[import]
from openai.lib.streaming.chat import (
    AsyncChatCompletionStreamManager as OpenAIAsyncChatCompletionStreamManager,
)
from openai.lib.streaming.chat import ChatCompletionStreamState
from openai.lib.streaming.chat import ChunkEvent as OpenAIChunkEvent
from pydantic import BaseModel

from ..cloud_llm import APIProvider, CloudLLM, CloudLLMSettings
from ..typing.tool import BaseTool
from . import (
    OpenAICompletion,
    OpenAICompletionChunk,
    OpenAIMessageParam,
    OpenAIParsedCompletion,
    OpenAIPredictionContentParam,
    OpenAIResponseFormatJSONObject,
    OpenAIResponseFormatText,
    OpenAIStreamOptionsParam,
    OpenAIToolChoiceOptionParam,
    OpenAIToolParam,
    OpenAIWebSearchOptions,
)
from .converters import OpenAIConverters

logger = logging.getLogger(__name__)


def get_openai_compatible_providers() -> list[APIProvider]:
    """Returns a dictionary of available OpenAI-compatible API providers."""
    return [
        APIProvider(
            name="openai",
            base_url="https://api.openai.com/v1",
            api_key=os.getenv("OPENAI_API_KEY"),
            response_schema_support=("*",),
        ),
        APIProvider(
            name="gemini_openai",
            base_url="https://generativelanguage.googleapis.com/v1beta/openai/",
            api_key=os.getenv("GEMINI_API_KEY"),
            response_schema_support=("*",),
        ),
        # Openrouter does not support structured outputs ATM
        APIProvider(
            name="openrouter",
            base_url="https://openrouter.ai/api/v1",
            api_key=os.getenv("OPENROUTER_API_KEY"),
            response_schema_support=(),
        ),
    ]


class OpenAILLMSettings(CloudLLMSettings, total=False):
    reasoning_effort: (
        Literal["none", "disable", "minimal", "low", "medium", "high"] | None
    )

    parallel_tool_calls: bool

    modalities: list[Literal["text"]] | None

    frequency_penalty: float | None
    presence_penalty: float | None
    logit_bias: dict[str, int] | None
    stop: str | list[str] | None
    logprobs: bool | None
    top_logprobs: int | None

    stream_options: OpenAIStreamOptionsParam | None

    prediction: OpenAIPredictionContentParam | None

    web_search_options: OpenAIWebSearchOptions | None

    metadata: dict[str, str] | None
    store: bool | None
    user: str

    # To support the old JSON mode without respose schemas
    response_format: OpenAIResponseFormatJSONObject | OpenAIResponseFormatText

    # TODO: support audio


@dataclass(frozen=True)
class OpenAILLM(CloudLLM):
    llm_settings: OpenAILLMSettings | None = None
    converters: ClassVar[OpenAIConverters] = OpenAIConverters()
    openai_client_timeout: float = 60.0
    openai_client_max_retries: int = 2
    extra_openai_client_params: dict[str, Any] | None = None
    client: AsyncOpenAI = field(init=False)

    def __post_init__(self):
        super().__post_init__()

        openai_compatible_providers = get_openai_compatible_providers()

        _api_provider = self.api_provider

        model_name_parts = self.model_name.split("/", 1)
        if _api_provider is not None:
            _model_name = self.model_name
        elif len(model_name_parts) == 2:
            compat_providers_map = {
                provider["name"]: provider for provider in openai_compatible_providers
            }
            _provider_name, _model_name = model_name_parts
            if _provider_name not in compat_providers_map:
                raise ValueError(
                    f"API provider '{_provider_name}' is not a supported OpenAI "
                    f"compatible provider. Supported providers are: "
                    f"{', '.join(compat_providers_map.keys())}"
                )
            _api_provider = compat_providers_map[_provider_name]
        else:
            raise ValueError(
                "Model name must be in the format 'provider/model_name' or "
                "you must provide an 'api_provider' argument."
            )

        response_schema_support: bool = any(
            fnmatch.fnmatch(_model_name, pat)
            for pat in _api_provider.get("response_schema_support") or []
        )
        if self.apply_response_schema_via_provider and not response_schema_support:
            raise ValueError(
                "Native response schema validation is not supported for model "
                f"'{_model_name}' by the API provider '{_api_provider['name']}'. "
                "Please set apply_response_schema_via_provider=False."
            )

        _openai_client_params = deepcopy(self.extra_openai_client_params or {})
        _openai_client_params["timeout"] = self.openai_client_timeout
        _openai_client_params["max_retries"] = self.openai_client_max_retries
        if self.http_client is not None:
            _openai_client_params["http_client"] = self.http_client

        _client = AsyncOpenAI(
            base_url=_api_provider.get("base_url"),
            api_key=_api_provider.get("api_key"),
            **_openai_client_params,
        )

        object.__setattr__(self, "model_name", _model_name)
        object.__setattr__(self, "api_provider", _api_provider)
        object.__setattr__(self, "client", _client)

    async def _get_api_completion(
        self,
        api_messages: Iterable[OpenAIMessageParam],
        *,
        api_tools: list[OpenAIToolParam] | None = None,
        api_tool_choice: OpenAIToolChoiceOptionParam | None = None,
        api_response_schema: type[Any] | None = None,
        **api_llm_settings: Any,
    ) -> OpenAICompletion | OpenAIParsedCompletion[Any]:
        tools = api_tools or NOT_GIVEN
        tool_choice = api_tool_choice or NOT_GIVEN
        response_format = api_response_schema or NOT_GIVEN

        if self.apply_response_schema_via_provider:
            return await self.client.beta.chat.completions.parse(
                model=self.model_name,
                messages=api_messages,
                tools=tools,
                tool_choice=tool_choice,
                response_format=response_format,
                **api_llm_settings,
            )

        return await self.client.chat.completions.create(
            model=self.model_name,
            messages=api_messages,
            tools=tools,
            tool_choice=tool_choice,
            stream=False,
            **api_llm_settings,
        )

    async def _get_api_completion_stream(
        self,
        api_messages: Iterable[OpenAIMessageParam],
        api_tools: list[OpenAIToolParam] | None = None,
        api_tool_choice: OpenAIToolChoiceOptionParam | None = None,
        api_response_schema: type[Any] | None = None,
        **api_llm_settings: Any,
    ) -> AsyncIterator[OpenAICompletionChunk]:
        tools = api_tools or NOT_GIVEN
        tool_choice = api_tool_choice or NOT_GIVEN
        response_format = api_response_schema or NOT_GIVEN

        # Ensure usage is included in the streamed responses
        stream_options = dict(api_llm_settings.get("stream_options") or {})
        stream_options["include_usage"] = True
        _api_llm_settings = api_llm_settings | {"stream_options": stream_options}

        # Need to wrap the iterator to make it work with decorators
        async def iterator() -> AsyncIterator[OpenAICompletionChunk]:
            if self.apply_response_schema_via_provider:
                stream_manager: OpenAIAsyncChatCompletionStreamManager[Any] = (
                    self.client.beta.chat.completions.stream(
                        model=self.model_name,
                        messages=api_messages,
                        tools=tools,
                        tool_choice=tool_choice,
                        response_format=response_format,
                        **_api_llm_settings,
                    )
                )
                async with stream_manager as stream:
                    async for chunk_event in stream:
                        if isinstance(chunk_event, OpenAIChunkEvent):
                            yield chunk_event.chunk
            else:
                stream_generator: AsyncStream[
                    OpenAICompletionChunk
                ] = await self.client.chat.completions.create(
                    model=self.model_name,
                    messages=api_messages,
                    tools=tools,
                    tool_choice=tool_choice,
                    stream=True,
                    **api_llm_settings,
                )
                async with stream_generator as stream:
                    async for completion_chunk in stream:
                        yield completion_chunk

        return iterator()

    def combine_completion_chunks(
        self,
        completion_chunks: list[OpenAICompletionChunk],
        response_schema: Any | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
    ) -> OpenAICompletion:
        response_format = NOT_GIVEN
        input_tools = NOT_GIVEN
        if self.apply_response_schema_via_provider and response_schema:
            response_format = response_schema
        if self.apply_tool_call_schema_via_provider and tools:
            input_tools = [
                self.converters.to_tool(tool, strict=True) for tool in tools.values()
            ]
        state = ChatCompletionStreamState[Any](
            input_tools=input_tools, response_format=response_format
        )
        for chunk in completion_chunks:
            state.handle_chunk(chunk)

        return state.get_final_completion()



import json
from collections.abc import Sequence
from typing import ClassVar, Generic, Protocol, TypeAlias, TypeVar, cast, final

from pydantic import BaseModel, TypeAdapter

from .errors import InputPromptBuilderError
from .generics_utils import AutoInstanceAttributesMixin
from .run_context import CtxT, RunContext
from .typing.content import Content, ImageData
from .typing.io import InT, LLMPrompt
from .typing.message import UserMessage
from .utils.callbacks import is_method_overridden

_InT_contra = TypeVar("_InT_contra", contravariant=True)


class SystemPromptBuilder(Protocol[CtxT]):
    def __call__(self, *, ctx: RunContext[CtxT], call_id: str) -> str | None: ...


class InputContentBuilder(Protocol[_InT_contra, CtxT]):
    def __call__(
        self, in_args: _InT_contra, *, ctx: RunContext[CtxT], call_id: str
    ) -> Content: ...


PromptArgumentType: TypeAlias = str | bool | int | ImageData


class PromptBuilder(AutoInstanceAttributesMixin, Generic[InT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {0: "_in_type"}

    def __init__(
        self, agent_name: str, sys_prompt: LLMPrompt | None, in_prompt: LLMPrompt | None
    ):
        self._in_type: type[InT]
        super().__init__()

        self._agent_name = agent_name
        self._sys_prompt = sys_prompt
        self._in_prompt = in_prompt
        self._in_args_type_adapter: TypeAdapter[InT] = TypeAdapter(self._in_type)

    @property
    def sys_prompt(self) -> LLMPrompt | None:
        return self._sys_prompt

    @property
    def in_prompt(self) -> LLMPrompt | None:
        return self._in_prompt

    @final
    def build_system_prompt(self, *, ctx: RunContext[CtxT], call_id: str) -> str | None:
        if is_method_overridden("build_system_prompt_impl", self):
            return self.build_system_prompt_impl(ctx=ctx, call_id=call_id)

        return self.sys_prompt

    @final
    def build_input_content(
        self, in_args: InT | None, *, ctx: RunContext[CtxT], call_id: str
    ) -> Content:
        if in_args is None and self._in_type is not type(None):
            raise InputPromptBuilderError(
                proc_name=self._agent_name,
                message="Either chat inputs or input arguments must be provided "
                f"when input type is not None [agent_name={self._agent_name}]",
            )

        in_args = cast("InT", in_args)
        val_in_args = self._validate_input_args(in_args)

        if is_method_overridden("build_input_content_impl", self):
            return self.build_input_content_impl(
                in_args=val_in_args, ctx=ctx, call_id=call_id
            )

        if issubclass(self._in_type, BaseModel) and isinstance(val_in_args, BaseModel):
            val_in_args_map = self._format_pydantic_prompt_args(val_in_args)
            if self.in_prompt is not None:
                return Content.from_formatted_prompt(self.in_prompt, **val_in_args_map)
            return Content.from_text(json.dumps(val_in_args_map, indent=2))

        fmt_in_args = self._in_args_type_adapter.dump_json(
            val_in_args, indent=2, warnings="error"
        ).decode("utf-8")
        return Content.from_text(fmt_in_args)

    @final
    def build_input_message(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_args: InT | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> UserMessage | None:
        if chat_inputs is not None:
            if in_args is not None:
                raise InputPromptBuilderError(
                    proc_name=self._agent_name,
                    message="Cannot use both chat inputs and input arguments "
                    f"at the same time [agent_name={self._agent_name}]",
                )
            if isinstance(chat_inputs, LLMPrompt):
                return UserMessage.from_text(chat_inputs, name=self._agent_name)
            return UserMessage.from_content_parts(chat_inputs, name=self._agent_name)

        return UserMessage(
            content=self.build_input_content(in_args=in_args, ctx=ctx, call_id=call_id),
            name=self._agent_name,
        )

    def _validate_input_args(self, in_args: InT) -> InT:
        val_in_args = self._in_args_type_adapter.validate_python(in_args)
        if isinstance(val_in_args, BaseModel):
            has_image = self._has_image_data(val_in_args)
            if has_image and self.in_prompt is None:
                raise InputPromptBuilderError(
                    proc_name=self._agent_name,
                    message="BaseModel input arguments contain ImageData, "
                    "but input prompt template is not set "
                    f"[agent_name={self._agent_name}]. Cannot format input arguments.",
                )
        elif self.in_prompt is not None:
            raise InputPromptBuilderError(
                proc_name=self._agent_name,
                message="Cannot use the input prompt template with "
                f"non-BaseModel input arguments [agent_name={self._agent_name}]",
            )

        return val_in_args

    @staticmethod
    def _has_image_data(inp: BaseModel) -> bool:
        contains_image_data = False
        for field in type(inp).model_fields:
            if isinstance(getattr(inp, field), ImageData):
                contains_image_data = True

        return contains_image_data

    @staticmethod
    def _format_pydantic_prompt_args(inp: BaseModel) -> dict[str, PromptArgumentType]:
        formatted_args: dict[str, PromptArgumentType] = {}
        for field_name, field_info in type(inp).model_fields.items():
            if field_info.exclude:
                continue

            val = getattr(inp, field_name)
            if isinstance(val, (int, str, bool, ImageData)):
                formatted_args[field_name] = val
            else:
                formatted_args[field_name] = (
                    TypeAdapter(type(val))  # type: ignore[return-value]
                    .dump_json(val, indent=2, warnings="error")
                    .decode("utf-8")
                )

        return formatted_args

    def build_system_prompt_impl(
        self, *, ctx: RunContext[CtxT], call_id: str
    ) -> str | None:
        raise NotImplementedError

    def build_input_content_impl(
        self, in_args: InT, *, ctx: RunContext[CtxT], call_id: str
    ) -> Content:
        raise NotImplementedError



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Literal,
    TypeAlias,
    TypeVar,
)

from pydantic import BaseModel, PrivateAttr, TypeAdapter

from grasp_agents.tracing_decorators import tool

from ..generics_utils import AutoInstanceAttributesMixin

if TYPE_CHECKING:
    from ..run_context import CtxT, RunContext
    from .events import Event
else:
    CtxT = TypeVar("CtxT")


_InT = TypeVar("_InT", bound=BaseModel)
_OutT_co = TypeVar("_OutT_co", covariant=True)


class ToolCall(BaseModel):
    id: str
    tool_name: str
    tool_arguments: str


@tool(name="tool", method_name="__call__")  # type: ignore
class BaseTool(
    AutoInstanceAttributesMixin,
    BaseModel,
    ABC,
    Generic[_InT, _OutT_co, CtxT],
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    name: str
    description: str

    _in_type: type[_InT] = PrivateAttr()
    _out_type: type[_OutT_co] = PrivateAttr()

    @property
    def in_type(self) -> type[_InT]:
        return self._in_type

    @property
    def out_type(self) -> type[_OutT_co]:
        return self._out_type

    @abstractmethod
    async def run(
        self,
        inp: _InT,
        *,
        ctx: RunContext[CtxT] | None = None,
        call_id: str | None = None,
    ) -> _OutT_co:
        pass

    async def run_stream(
        self,
        inp: _InT,
        *,
        ctx: RunContext[CtxT] | None = None,
        call_id: str | None = None,
    ) -> AsyncIterator[Event[Any]]:
        from .events import ToolOutputEvent

        out = await self.run(inp, ctx=ctx, call_id=call_id)
        yield ToolOutputEvent(data=out, src_name=self.name, call_id=call_id)

    async def __call__(
        self,
        *,
        ctx: RunContext[CtxT] | None = None,
        call_id: str | None = None,
        **kwargs: Any,
    ) -> _OutT_co:
        # NOTE: validation is probably redundant here when tool inputs have been
        # validated by the LLM already
        input_args = TypeAdapter(self._in_type).validate_python(kwargs)
        output = await self.run(input_args, ctx=ctx, call_id=call_id)

        return TypeAdapter(self._out_type).validate_python(output)


class NamedToolChoice(BaseModel):
    name: str


ToolChoice: TypeAlias = Literal["none", "auto", "required"] | NamedToolChoice



from __future__ import annotations

import asyncio
import contextlib
import functools
from collections.abc import AsyncGenerator, AsyncIterable, AsyncIterator, Callable
from logging import getLogger
from typing import Any, Generic, ParamSpec, TypeVar

from grasp_agents.typing.events import Event

logger = getLogger(__name__)


_T = TypeVar("_T")
_K = TypeVar("_K")
_P = ParamSpec("_P")


async def stream_concurrent(
    generators: list[AsyncIterator[_T]],
) -> AsyncIterator[tuple[int, _T]]:
    tasks: list[asyncio.Task[None]] = []
    queue: asyncio.Queue[tuple[int, _T] | None] = asyncio.Queue()
    pumps_left = len(generators)

    async def pump(gen: AsyncIterator[_T], idx: int) -> None:
        nonlocal pumps_left
        try:
            async for item in gen:
                await queue.put((idx, item))

        except asyncio.CancelledError:
            raise

        except Exception as e:
            logger.warning(
                f"stream_concurrent pump {idx} failed:\n{e!r}", exc_info=True
            )

        finally:
            pumps_left -= 1
            if pumps_left == 0:
                await queue.put(None)

    async with asyncio.TaskGroup() as tg:
        for idx, gen in enumerate(generators):
            tasks.append(tg.create_task(pump(gen, idx)))

        while True:
            msg = await queue.get()
            if msg is None:
                break
            yield msg


_F = TypeVar("_F")


class MissingFinalEventError(RuntimeError):
    """Raised when the stream finishes without encountering the required final event type."""


class EventStream(AsyncIterator[Event[Any]], Generic[_F]):
    def __init__(
        self, source: AsyncIterable[Event[Any]], final_type: type[_F] = object
    ) -> None:
        self._aiter: AsyncIterator[Event[Any]] = source.__aiter__()
        self._final_type: type[_F] = final_type
        self._final_event: Event[_F]
        self._final_event_set: bool = False
        self._events: list[Event[Any]] = []

    @property
    def final_type(self) -> type[_F]:
        return self._final_type

    @property
    def events(self) -> list[Event[Any]]:
        return self._events

    def __aiter__(self) -> EventStream[_F]:
        return self

    async def __anext__(self) -> Event[Any]:
        event = await self._aiter.__anext__()
        if isinstance(event.data, self.final_type):
            self._final_event = event
            self._final_event_set = True

        return event

    async def drain(self) -> list[Event[Any]]:
        async for event in self:
            self._events.append(event)
        return self._events

    async def final_event(self) -> Event[_F]:
        async for _ in self:
            pass
        if not self._final_event_set:
            raise MissingFinalEventError(
                f"No event of type Event[{self.final_type.__name__}] was encountered."
            )
        return self._final_event

    async def final_data(self) -> _F:
        return (await self.final_event()).data


def wrap_with_aclosing(
    async_func: Callable[_P, AsyncGenerator[_K]],
) -> Callable[_P, AsyncGenerator[_K]]:
    @functools.wraps(async_func)
    async def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> AsyncGenerator[_K]:
        async with contextlib.aclosing(async_func(*args, **kwargs)) as stream:
            async for obj in stream:
                yield obj

    return wrapper



import asyncio
import logging
from collections.abc import AsyncIterator, Sequence
from itertools import chain
from typing import Any

from grasp_agents.utils.streaming import stream_concurrent

from ..errors import ProcInputValidationError
from ..packet import Packet
from ..run_context import CtxT, RunContext
from ..typing.events import Event, ProcPacketOutEvent, ProcPayloadOutEvent
from ..typing.io import InT, OutT, ProcName
from .processor import Processor

logger = logging.getLogger(__name__)


class ParallelProcessor(Processor[InT, OutT, CtxT]):
    def __init__(self, subproc: Processor[InT, OutT, CtxT]) -> None:
        super().__init__(
            name=subproc.name + "_par", recipients=subproc.recipients, max_retries=0
        )

        self._in_type = subproc.in_type
        self._out_type = subproc.out_type
        self._subproc = subproc

        # This disables recipient selection in the subprocessor,
        # but preserves subproc.select_recipients_impl
        subproc.recipients = None

    def select_recipients_impl(
        self, output: OutT, *, ctx: RunContext[CtxT], call_id: str
    ) -> Sequence[ProcName]:
        # Move recipient selection to the outer ParallelProcessor
        return self._subproc.select_recipients_impl(
            output=output, ctx=ctx, call_id=call_id
        )

    @property
    def subproc(self) -> Processor[InT, OutT, CtxT]:
        return self._subproc

    def _validate_in_args(
        self,
        chat_inputs: Any | None = None,
        in_args: list[InT] | None = None,
        *,
        call_id: str,
    ) -> list[InT]:
        err_kwargs = {"proc_name": self.name, "call_id": call_id}
        if chat_inputs is not None:
            raise ProcInputValidationError(
                message=f"ParallelProcessor {self.name} does not support chat_inputs",
                **err_kwargs,
            )
        if in_args is None:
            raise ProcInputValidationError(
                message=f"ParallelProcessor {self.name} requires in_args to be provided",
                **err_kwargs,
            )

        return in_args

    def _join_payloads_from_packets(
        self, packets: Sequence[Packet[OutT]]
    ) -> list[OutT]:
        return list(
            Packet(
                payloads=list(chain.from_iterable(p.payloads for p in packets)),
                sender=self.name,
            ).payloads
        )

    async def _process(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> list[OutT]:
        in_args = self._validate_in_args(
            chat_inputs=chat_inputs, in_args=in_args, call_id=call_id
        )
        subproc_replicas = [self._subproc.copy() for _ in in_args]
        corouts = [
            proc.run(in_args=inp, call_id=f"{call_id}/{idx}", ctx=ctx)
            for idx, (inp, proc) in enumerate(
                zip(in_args, subproc_replicas, strict=True)
            )
        ]
        out_packets = await asyncio.gather(*corouts)

        return self._join_payloads_from_packets(out_packets)

    async def _process_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> AsyncIterator[Event[Any]]:
        in_args = self._validate_in_args(
            chat_inputs=chat_inputs, in_args=in_args, call_id=call_id
        )
        subproc_replicas = [self._subproc.copy() for _ in in_args]

        streams = [
            proc.run_stream(in_args=inp, call_id=f"{call_id}/{idx}", ctx=ctx)
            for idx, (inp, proc) in enumerate(
                zip(in_args, subproc_replicas, strict=True)
            )
        ]
        out_packets_map: dict[int, Packet[OutT]] = {}
        async for idx, event in stream_concurrent(streams):
            if (
                isinstance(event, ProcPacketOutEvent)
                and event.src_name == self._subproc.name
            ):
                out_packets_map[idx] = event.data
            else:
                yield event

        out_packets = [out_packets_map[idx] for idx in sorted(out_packets_map)]

        # Need to emit ProcPayloadOutputEvent in the order of in_args,
        # thus we filter them out first, then yield in order.

        for p in self._join_payloads_from_packets(out_packets):
            yield ProcPayloadOutEvent(data=p, src_name=self.name, call_id=call_id)



import contextlib
import inspect
import json
import os
import types
import warnings
from collections.abc import Awaitable, Callable
from functools import wraps
from typing import (
    Any,
    ParamSpec,
    TypeVar,
    cast,
)

from opentelemetry import context as context_api
from opentelemetry import trace
from opentelemetry.context import _RUNTIME_CONTEXT
from opentelemetry.semconv_ai import SpanAttributes, TraceloopSpanKindValues
from opentelemetry.trace.status import Status, StatusCode
from pydantic import BaseModel
from traceloop.sdk.telemetry import Telemetry
from traceloop.sdk.tracing import get_tracer, set_workflow_name
from traceloop.sdk.tracing.tracing import (
    TracerWrapper,
    get_chained_entity_path,
    set_entity_path,
)
from traceloop.sdk.utils import camel_to_snake
from traceloop.sdk.utils.json_encoder import JSONEncoder

P = ParamSpec("P")

R = TypeVar("R")
F = TypeVar("F", bound=Callable[P, R | Awaitable[R]])

DEFAULT_EXCLUDE_FIELDS = {"_hidden_params", "completions"}


def _to_plain(obj: Any, exclude_fields: set[str] | None = None) -> Any:
    """
    Recursively convert objects to JSON-serializable primitives.

    - Pydantic BaseModel -> dict via model_dump()
    - dict/list/tuple/set -> recurse (sets become lists)
    - other objects -> returned as-is (left to JSONEncoder)
    """
    if isinstance(obj, BaseModel):
        try:
            # TODO: remove this temporary hack
            return obj.model_dump(
                exclude=DEFAULT_EXCLUDE_FIELDS.union(exclude_fields or set())
            )
        except Exception:
            return str(obj)
    if isinstance(obj, dict):
        items_dict = cast("dict[Any, Any]", obj)
        result: dict[str, Any] = {}
        for k, v in items_dict.items():
            result[str(k)] = _to_plain(v)
        return result
    if isinstance(obj, (tuple, list, set)):
        items = cast("tuple[Any, ...] | list[Any] | set[Any]", obj)
        lst: list[Any] = []
        for v in items:
            lst.append(_to_plain(v))
        return lst
    return obj


def _truncate_json_if_needed(json_str: str) -> str:
    """
    Truncate JSON string if it exceeds OTEL_SPAN_ATTRIBUTE_VALUE_LENGTH_LIMIT;
    truncation may yield an invalid JSON string, which is expected for logging purposes.
    """
    limit_str = os.getenv("OTEL_SPAN_ATTRIBUTE_VALUE_LENGTH_LIMIT")
    if limit_str:
        try:
            limit = int(limit_str)
            if limit > 0 and len(json_str) > limit:
                return json_str[:limit]
        except ValueError:
            pass
    return json_str


def _handle_generator(span, res):
    # for some reason the SPAN_KEY is not being set in the context of the generator, so we re-set it
    context_api.attach(trace.set_span_in_context(span))
    try:
        for item in res:
            yield item
    except Exception as e:
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.record_exception(e)
        raise
    finally:
        span.end()
        # Note: we don't detach the context here as this fails in some situations
        # https://github.com/open-telemetry/opentelemetry-python/issues/2606
        # This is not a problem since the context will be detached automatically during garbage collection


async def _ahandle_generator(span, ctx_token, res):
    try:
        async for part in res:
            yield part
    except Exception as e:
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.record_exception(e)
        raise
    finally:
        span.end()
        context_api.detach(ctx_token)


def _should_send_prompts():
    return (
        os.getenv("TRACELOOP_TRACE_CONTENT") or "true"
    ).lower() == "true" or context_api.get_value("override_enable_content_tracing")


# Quiet wrapper that suppresses prints and warnings from TracerWrapper.verify_initialized
def _tracing_initialized_quietly() -> bool:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            with (
                open(os.devnull, "w") as devnull,
                contextlib.redirect_stdout(devnull),
                contextlib.redirect_stderr(devnull),
            ):
                return TracerWrapper.verify_initialized()
        except Exception:
            return False


# Unified Decorators : handles both sync and async operations


def _is_async_method(fn):
    # check if co-routine function or async generator( example : using async & yield)
    return inspect.iscoroutinefunction(fn) or inspect.isasyncgenfunction(fn)


def _setup_span(
    entity_name,
    tlp_span_kind,
    version,
    # --- added by Grasp ---
    instance: type | None = None,
    kwargs: dict[str, Any] | None = None,
    # ---
):
    """Sets up the OpenTelemetry span and context"""
    if tlp_span_kind in [
        TraceloopSpanKindValues.WORKFLOW,
        TraceloopSpanKindValues.AGENT,
    ]:
        set_workflow_name(entity_name)

    # --- added by Grasp ---

    instance_name = None

    if instance is not None:
        if tlp_span_kind in [
            TraceloopSpanKindValues.WORKFLOW,
            TraceloopSpanKindValues.AGENT,
            TraceloopSpanKindValues.TOOL,
        ]:
            instance_name = getattr(instance, "name", None)

        elif (
            tlp_span_kind == TraceloopSpanKindValues.TASK and entity_name == "generate"
        ):
            instance_name = getattr(instance, "agent_name", None)

    if instance_name:
        call_id = (kwargs or {}).get("call_id")
        span_name = f"{instance_name}.{entity_name}" + (
            f"[{call_id}]" if call_id else ""
        )
    else:
        # ---
        span_name = f"{entity_name}.{tlp_span_kind.value}"

    with get_tracer() as tracer:
        span = tracer.start_span(span_name)
        ctx = trace.set_span_in_context(span)
        ctx_token = context_api.attach(ctx)

        if tlp_span_kind in [
            TraceloopSpanKindValues.TASK,
            TraceloopSpanKindValues.TOOL,
        ]:
            entity_path = get_chained_entity_path(entity_name)
            set_entity_path(entity_path)

        span.set_attribute(SpanAttributes.TRACELOOP_SPAN_KIND, tlp_span_kind.value)
        span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_NAME, entity_name)
        if version:
            span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_VERSION, version)

    return span, ctx, ctx_token


def _handle_span_input(
    span, args, kwargs, cls=None, exclude_fields: set[str] | None = None
):
    """Handles entity input logging in JSON for both sync and async functions"""
    try:
        if _should_send_prompts():
            json_input = json.dumps(
                {
                    "args": _to_plain(list(args), exclude_fields=exclude_fields),
                    "kwargs": _to_plain(kwargs, exclude_fields=exclude_fields),
                },
                **({"cls": cls} if cls else {}),
                indent=2,
            )  # JSON formatting updated by Grasp
            truncated_json = _truncate_json_if_needed(json_input)
            span.set_attribute(
                SpanAttributes.TRACELOOP_ENTITY_INPUT,
                truncated_json,
            )
    except TypeError as e:
        Telemetry().log_exception(e)


def _handle_span_output(span, res, cls=None, exclude_fields: set[str] | None = None):
    """Handles entity output logging in JSON for both sync and async functions"""
    try:
        if _should_send_prompts():
            json_output = json.dumps(
                _to_plain(res, exclude_fields=exclude_fields),
                **({"cls": cls} if cls else {}),
                indent=2,
            )  # JSON formatting updated by Grasp
            truncated_json = _truncate_json_if_needed(json_output)
            span.set_attribute(
                SpanAttributes.TRACELOOP_ENTITY_OUTPUT,
                truncated_json,
            )
    except TypeError as e:
        Telemetry().log_exception(e)


def _cleanup_span(span, ctx_token):
    """End the span process and detach the context token"""
    span.end()
    try:
        _RUNTIME_CONTEXT.detach(ctx_token)
    except Exception:
        pass


# --- added by Grasp ---
def is_bound_method(func: Callable[..., Any], self_candidate: Any) -> bool:
    return (inspect.ismethod(func) and (func.__self__ is self_candidate)) or hasattr(
        self_candidate, func.__name__
    )


# ---


def entity_method(
    name: str | None = None,
    version: int | None = None,
    tlp_span_kind: TraceloopSpanKindValues | None = TraceloopSpanKindValues.TASK,
) -> Callable[[F], F]:
    def decorate(fn: F) -> F:
        is_async = _is_async_method(fn)
        entity_name = name or fn.__qualname__
        if is_async:
            if inspect.isasyncgenfunction(fn):

                @wraps(fn)
                async def async_gen_wrap(*args: Any, **kwargs: Any) -> Any:
                    if not _tracing_initialized_quietly():
                        async for item in fn(*args, **kwargs):
                            yield item
                        return

                    # --- added by Grasp ---
                    is_bound = is_bound_method(fn, args[0] if args else False)
                    instance = args[0] if is_bound else None
                    # ---

                    span, ctx, ctx_token = _setup_span(
                        entity_name,
                        tlp_span_kind,
                        version,
                        # --- added by Grasp ---
                        instance=instance,
                        kwargs=kwargs,
                        # ---
                    )
                    _handle_span_input(span, args, kwargs, cls=JSONEncoder)

                    #  --- added by Grasp ---
                    items = []
                    try:
                        async for item in fn(*args, **kwargs):
                            # span.add_event(
                            #     name=getattr(item, "type", "unknown"),
                            #     attributes=_to_plain(item),
                            # )
                            items.append(item)
                            yield item
                    except Exception as e:
                        span.set_status(Status(StatusCode.ERROR, str(e)))
                        span.record_exception(e)
                        raise
                    finally:
                        if items:
                            _handle_span_output(span, items[-1], cls=JSONEncoder)

                        # Detaching context for generators may cause issues:
                        # https://github.com/open-telemetry/opentelemetry-python/issues/2606
                        _cleanup_span(span, ctx_token)
                        # span.end()

                    #  ---

                    # async for item in _ahandle_generator(
                    #     span, ctx_token, fn(*args, **kwargs)
                    # ):
                    #     yield item

                return cast("F", async_gen_wrap)

            @wraps(fn)
            async def async_wrap(*args: Any, **kwargs: Any) -> Any:
                if not _tracing_initialized_quietly():
                    return await fn(*args, **kwargs)

                # --- added by Grasp ---
                is_bound = is_bound_method(fn, args[0] if args else False)
                instance = args[0] if is_bound else None
                # ---

                span, ctx, ctx_token = _setup_span(
                    entity_name,
                    tlp_span_kind,
                    version,
                    # --- added by Grasp ---
                    instance=instance,
                    kwargs=kwargs,
                    # ---
                )
                _handle_span_input(span, args, kwargs, cls=JSONEncoder)
                try:
                    res = await fn(*args, **kwargs)
                    _handle_span_output(span, res, cls=JSONEncoder)
                    return res
                except Exception as e:
                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    span.record_exception(e)
                    raise
                finally:
                    _cleanup_span(span, ctx_token)

            return cast("F", async_wrap)

        @wraps(fn)
        def sync_wrap(*args: Any, **kwargs: Any) -> Any:
            if not _tracing_initialized_quietly():
                return fn(*args, **kwargs)

            # --- added by Grasp ---
            is_bound = is_bound_method(fn, args[0] if args else False)
            instance = args[0] if is_bound else None
            # ---

            span, ctx, ctx_token = _setup_span(
                entity_name,
                tlp_span_kind,
                version,
                # --- added by Grasp ---
                instance=instance,
                kwargs=kwargs,
                # ---
            )
            _handle_span_input(span, args, kwargs, cls=JSONEncoder)
            try:
                res = fn(*args, **kwargs)
            except Exception as e:
                span.set_status(Status(StatusCode.ERROR, str(e)))
                span.record_exception(e)
                _cleanup_span(span, ctx_token)
                raise

            if isinstance(res, types.GeneratorType):
                # --- added by Grasp ---
                context_api.attach(trace.set_span_in_context(span))
                items = []
                try:
                    for item in res:
                        items.append(item)
                        yield item
                except Exception as e:
                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    span.record_exception(e)
                    raise
                finally:
                    if items:
                        _handle_span_output(span, items[-1], cls=JSONEncoder)

                    # Detaching context for generators may cause issues:
                    # https://github.com/open-telemetry/opentelemetry-python/issues/2606
                    _cleanup_span(span, ctx_token)
                    # span.end()
                return
                # ---

                # _handle_generator(span, res)

            _handle_span_output(span, res, cls=JSONEncoder)
            _cleanup_span(span, ctx_token)
            return res

        return cast("F", sync_wrap)

    return decorate


def entity_class(
    name: str | None,
    version: int | None,
    method_name: str,
    tlp_span_kind: TraceloopSpanKindValues | None = TraceloopSpanKindValues.TASK,
):
    def decorator(cls):
        task_name = name or camel_to_snake(cls.__qualname__)
        method = getattr(cls, method_name)
        setattr(
            cls,
            method_name,
            entity_method(name=task_name, version=version, tlp_span_kind=tlp_span_kind)(
                method
            ),
        )
        return cls

    return decorator



from collections.abc import AsyncIterator, Mapping, Sequence
from copy import deepcopy
from pathlib import Path
from typing import Any, ClassVar, Generic, Protocol, TypedDict, TypeVar, cast, final

from pydantic import BaseModel

from .errors import ProcInputValidationError
from .llm import LLM
from .llm_agent_memory import LLMAgentMemory
from .llm_policy_executor import (
    AfterGenerateHook,
    BeforeGenerateHook,
    FinalAnswerChecker,
    LLMPolicyExecutor,
    ToolOutputConverter,
)
from .processors.processor import Processor
from .prompt_builder import (
    InputContentBuilder,
    PromptBuilder,
    SystemPromptBuilder,
)
from .run_context import CtxT, RunContext
from .typing.content import Content, ImageData
from .typing.events import (
    Event,
    ProcPayloadOutEvent,
    SystemMessageEvent,
    UserMessageEvent,
)
from .typing.io import InT, LLMPrompt, OutT, ProcName
from .typing.message import AssistantMessage, Message, SystemMessage, UserMessage
from .typing.tool import BaseTool, ToolCall
from .utils.callbacks import is_method_overridden
from .utils.io import get_prompt
from .utils.validation import validate_obj_from_json_or_py_string

_InT_contra = TypeVar("_InT_contra", contravariant=True)
_OutT_co = TypeVar("_OutT_co", covariant=True)


class MemoryPreparator(Protocol[_InT_contra]):
    def __call__(
        self,
        *,
        instructions: LLMPrompt | None = None,
        in_args: _InT_contra | None = None,
        ctx: RunContext[Any],
        call_id: str,
    ) -> None: ...


class OutputParser(Protocol[_InT_contra, _OutT_co, CtxT]):
    def __call__(
        self,
        final_answer: str,
        *,
        in_args: _InT_contra | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> _OutT_co: ...


class CallArgs(TypedDict):
    ctx: RunContext[Any]
    call_id: str


class LLMAgent(Processor[InT, OutT, CtxT], Generic[InT, OutT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        name: ProcName,
        *,
        # LLM
        llm: LLM,
        # Tools
        tools: list[BaseTool[Any, Any, CtxT]] | None = None,
        # Input prompt template (combines user and received arguments)
        in_prompt: LLMPrompt | None = None,
        in_prompt_path: str | Path | None = None,
        # System prompt template
        sys_prompt: LLMPrompt | None = None,
        sys_prompt_path: str | Path | None = None,
        # LLM response validation
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        # Agent loop settings
        max_turns: int = 100,
        react_mode: bool = False,
        final_answer_as_tool_call: bool = False,
        # Agent memory management
        memory: LLMAgentMemory | None = None,
        reset_memory_on_run: bool = False,
        # Agent run retries
        max_retries: int = 0,
        # Multi-agent routing
        recipients: Sequence[ProcName] | None = None,
        # Streaming
        stream_llm_responses: bool = False,
        stream_tools: bool = False,
    ) -> None:
        super().__init__(
            name=name, memory=memory, recipients=recipients, max_retries=max_retries
        )

        # Memory

        # Avoid narrowing the base '_memory' type (declared as 'Memory' in BaseProcessor)
        self._memory = memory or LLMAgentMemory()
        self._reset_memory_on_run = reset_memory_on_run

        # Prompt builder

        sys_prompt = get_prompt(prompt_text=sys_prompt, prompt_path=sys_prompt_path)
        in_prompt = get_prompt(prompt_text=in_prompt, prompt_path=in_prompt_path)

        self._prompt_builder = PromptBuilder[self.in_type, CtxT](
            agent_name=self._name, sys_prompt=sys_prompt, in_prompt=in_prompt
        )

        # LLM policy executor

        if issubclass(self._out_type, BaseModel):
            final_answer_type = self._out_type
        elif not final_answer_as_tool_call:
            final_answer_type = BaseModel
        else:
            raise TypeError(
                "Final answer type must be a subclass of BaseModel if "
                "final_answer_as_tool_call is True."
            )

        self._used_default_llm_response_schema: bool = False
        if (
            response_schema is None
            and tools is None
            and not is_method_overridden(
                "parse_output_impl", self, LLMAgent[Any, Any, Any]
            )
        ):
            response_schema = self.out_type
            self._used_default_llm_response_schema = True

        self._policy_executor: LLMPolicyExecutor[CtxT] = LLMPolicyExecutor[CtxT](
            agent_name=self.name,
            llm=llm,
            tools=tools,
            memory=self.memory,
            response_schema=response_schema,
            response_schema_by_xml_tag=response_schema_by_xml_tag,
            max_turns=max_turns,
            react_mode=react_mode,
            final_answer_type=final_answer_type,
            final_answer_as_tool_call=final_answer_as_tool_call,
            stream_llm_responses=stream_llm_responses,
            stream_tools=stream_tools,
        )

        self._register_overridden_implementations()

    @property
    def llm(self) -> LLM:
        return self._policy_executor.llm

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, CtxT]]:
        return self._policy_executor.tools

    @property
    def max_turns(self) -> int:
        return self._policy_executor.max_turns

    @property
    def sys_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.sys_prompt

    @property
    def in_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.in_prompt

    @property
    def memory(self) -> LLMAgentMemory:
        return cast("LLMAgentMemory", self._memory)

    @property
    def reset_memory_on_run(self) -> bool:
        return self._reset_memory_on_run

    @final
    def prepare_memory(
        self,
        *,
        instructions: LLMPrompt | None = None,
        in_args: InT | None = None,
        ctx: RunContext[Any],
        call_id: str,
    ) -> None:
        if is_method_overridden("prepare_memory_impl", self, LLMAgent[Any, Any, Any]):
            return self.prepare_memory_impl(
                instructions=instructions,
                in_args=in_args,
                ctx=ctx,
                call_id=call_id,
            )

    def _memorize_inputs(
        self,
        *,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        in_args: InT | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> tuple[SystemMessage | None, UserMessage | None]:
        call_kwargs = CallArgs(ctx=ctx, call_id=call_id)

        formatted_sys_prompt = self._prompt_builder.build_system_prompt(
            ctx=ctx, call_id=call_id
        )

        system_message: SystemMessage | None = None
        if self._reset_memory_on_run or self.memory.is_empty:
            self.memory.reset(formatted_sys_prompt)
            if formatted_sys_prompt is not None:
                system_message = cast("SystemMessage", self.memory.messages[0])
        else:
            self.prepare_memory(
                instructions=formatted_sys_prompt, in_args=in_args, **call_kwargs
            )

        input_message = self._prompt_builder.build_input_message(
            chat_inputs=chat_inputs, in_args=in_args, **call_kwargs
        )
        if input_message:
            self.memory.update([input_message])

        if system_message:
            self._print_messages([system_message], **call_kwargs)
        if input_message:
            self._print_messages([input_message], **call_kwargs)

        return system_message, input_message

    def parse_output_default(self, final_answer: str) -> OutT:
        return validate_obj_from_json_or_py_string(
            final_answer,
            schema=self._out_type,
            from_substring=False,
            strip_language_markdown=True,
        )

    @final
    def parse_output(
        self,
        final_answer: str,
        *,
        in_args: InT | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> OutT:
        if is_method_overridden("parse_output_impl", self, LLMAgent[Any, Any, Any]):
            return self.parse_output_impl(
                final_answer,
                in_args=in_args,
                ctx=ctx,
                call_id=call_id,
            )

        return self.parse_output_default(final_answer)

    def _extract_input_args(
        self, in_args: list[InT] | None, call_id: str
    ) -> InT | None:
        if in_args and len(in_args) != 1:
            raise ProcInputValidationError(
                proc_name=self.name,
                call_id=call_id,
                message="LLMAgent expects a single input argument.",
            )

        return in_args[0] if in_args else None

    async def _process_stream(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_args: list[InT] | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> AsyncIterator[Event[Any]]:
        call_kwargs = CallArgs(ctx=ctx, call_id=call_id)

        inp = self._extract_input_args(in_args, call_id)
        system_message, input_message = self._memorize_inputs(
            chat_inputs=chat_inputs, in_args=inp, **call_kwargs
        )
        if system_message:
            yield SystemMessageEvent(
                data=system_message, src_name=self.name, call_id=call_id
            )
        if input_message:
            yield UserMessageEvent(
                data=input_message, src_name=self.name, call_id=call_id
            )

        async for event in self._policy_executor.execute_stream(**call_kwargs):
            yield event
        final_answer = self._policy_executor.get_final_answer()

        output = self.parse_output(final_answer or "", in_args=inp, **call_kwargs)
        yield ProcPayloadOutEvent(data=output, src_name=self.name, call_id=call_id)

    async def _process(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_args: list[InT] | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> list[OutT]:
        async for event in self._process_stream(
            chat_inputs=chat_inputs, in_args=in_args, ctx=ctx, call_id=call_id
        ):
            if isinstance(event, ProcPayloadOutEvent):
                return [event.data]
        return []

    def _print_messages(
        self, messages: Sequence[Message], ctx: RunContext[CtxT], call_id: str
    ) -> None:
        if ctx.printer:
            ctx.printer.print_messages(messages, agent_name=self.name, call_id=call_id)

    # Methods that can be overridden in subclasses

    def prepare_memory_impl(
        self,
        *,
        instructions: LLMPrompt | None = None,
        in_args: InT | None = None,
        ctx: RunContext[Any],
        call_id: str,
    ) -> None:
        raise NotImplementedError

    def parse_output_impl(
        self,
        final_answer: str,
        *,
        in_args: InT | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> OutT:
        raise NotImplementedError

    def build_system_prompt_impl(
        self, *, ctx: RunContext[CtxT], call_id: str
    ) -> str | None:
        return self._prompt_builder.build_system_prompt_impl(ctx=ctx, call_id=call_id)

    def build_input_content_impl(
        self, in_args: InT, *, ctx: RunContext[CtxT], call_id: str
    ) -> Content:
        return self._prompt_builder.build_input_content_impl(
            in_args=in_args, ctx=ctx, call_id=call_id
        )

    def check_for_final_answer_impl(
        self,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> str | None:
        return self._policy_executor.check_for_final_answer_impl(
            ctx=ctx, call_id=call_id, **kwargs
        )

    async def on_before_generate_impl(
        self,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        extra_llm_settings: dict[str, Any],
    ) -> None:
        return await self._policy_executor.on_before_generate_impl(
            ctx=ctx,
            call_id=call_id,
            num_turns=num_turns,
            extra_llm_settings=extra_llm_settings,
        )

    async def on_after_generate_impl(
        self,
        gen_message: AssistantMessage,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
    ) -> None:
        return await self._policy_executor.on_after_generate_impl(
            gen_message=gen_message,
            ctx=ctx,
            call_id=call_id,
            num_turns=num_turns,
        )

    def tool_outputs_to_messages_impl(
        self,
        tool_outputs: Sequence[Any],
        tool_calls: Sequence[ToolCall],
        *,
        ctx: RunContext[CtxT],
        call_id: str,
    ):
        return self._policy_executor.tool_outputs_to_messages_impl(
            tool_outputs=tool_outputs,
            tool_calls=tool_calls,
            ctx=ctx,
            call_id=call_id,
        )

    # Decorators as an alternative to overriding methods

    def add_output_parser(
        self, func: OutputParser[InT, OutT, CtxT]
    ) -> OutputParser[InT, OutT, CtxT]:
        if self._used_default_llm_response_schema:
            self._policy_executor.response_schema = None
        self.parse_output_impl = func
        return func

    def add_memory_preparator(
        self, func: MemoryPreparator[InT]
    ) -> MemoryPreparator[InT]:
        self.prepare_memory_impl = func
        return func

    def add_system_prompt_builder(
        self, func: SystemPromptBuilder[CtxT]
    ) -> SystemPromptBuilder[CtxT]:
        self._prompt_builder.build_system_prompt_impl = func
        return func

    def add_input_content_builder(
        self, func: InputContentBuilder[InT, CtxT]
    ) -> InputContentBuilder[InT, CtxT]:
        self._prompt_builder.build_input_content_impl = func
        return func

    def add_final_answer_checker(
        self, func: FinalAnswerChecker[CtxT]
    ) -> FinalAnswerChecker[CtxT]:
        self._policy_executor.check_for_final_answer_impl = func
        return func

    def add_before_generate_hook(
        self, func: BeforeGenerateHook[CtxT]
    ) -> BeforeGenerateHook[CtxT]:
        self._policy_executor.on_before_generate_impl = func
        return func

    def add_after_generate_hook(
        self, func: AfterGenerateHook[CtxT]
    ) -> AfterGenerateHook[CtxT]:
        self._policy_executor.on_after_generate_impl = func
        return func

    def add_tool_output_converter(
        self, func: ToolOutputConverter[CtxT]
    ) -> ToolOutputConverter[CtxT]:
        self._policy_executor.tool_outputs_to_messages_impl = func
        return func

    # When methods are overridden in subclasses, pass them to the components

    def _register_overridden_implementations(self) -> None:
        base_cls = LLMAgent[Any, Any, Any]

        # Prompt builder

        if is_method_overridden("build_system_prompt_impl", self, base_cls):
            self._prompt_builder.build_system_prompt_impl = (
                self.build_system_prompt_impl
            )

        if is_method_overridden("build_input_content_impl", self, base_cls):
            self._prompt_builder.build_input_content_impl = (
                self.build_input_content_impl
            )

        # Policy executor

        if is_method_overridden("check_for_final_answer_impl", self, base_cls):
            self._policy_executor.check_for_final_answer_impl = (
                self.check_for_final_answer_impl
            )

        if is_method_overridden("on_before_generate_impl", self, base_cls):
            self._policy_executor.on_before_generate_impl = self.on_before_generate_impl

        if is_method_overridden("on_after_generate_impl", self, base_cls):
            self._policy_executor.on_after_generate_impl = self.on_after_generate_impl

        if is_method_overridden("tool_outputs_to_messages_impl", self, base_cls):
            self._policy_executor.tool_outputs_to_messages_impl = (
                self.tool_outputs_to_messages_impl
            )

    def copy(self) -> "LLMAgent[InT, OutT, CtxT]":
        # Share LLM and tools, deepcopy everything else

        cls = self.__class__
        new_obj = cls.__new__(cls)
        memo = {id(self): new_obj}

        pe = getattr(self, "_policy_executor", None)
        if pe is not None:
            if getattr(pe, "llm", None) is not None:
                memo[id(pe.llm)] = pe.llm
            if getattr(pe, "tools", None) is not None:
                memo[id(pe.tools)] = pe.tools

        state = deepcopy(self.__dict__, memo)
        new_obj.__dict__.update(state)

        return new_obj



# pyright: reportUnusedImport=false


from .llm import LLM, LLMSettings
from .llm_agent import LLMAgent
from .llm_agent_memory import LLMAgentMemory
from .memory import Memory
from .packet import Packet
from .printer import Printer, print_event_stream
from .processors.base_processor import BaseProcessor
from .processors.parallel_processor import ParallelProcessor
from .processors.processor import Processor
from .run_context import RunContext
from .typing.completion import Completion
from .typing.content import Content, ImageData
from .typing.io import LLMPrompt, ProcName
from .typing.message import AssistantMessage, Messages, SystemMessage, UserMessage
from .typing.tool import BaseTool

__all__ = [
    "LLM",
    "AssistantMessage",
    "BaseProcessor",
    "BaseTool",
    "Completion",
    "Content",
    "ImageData",
    "LLMAgent",
    "LLMAgentMemory",
    "LLMPrompt",
    "LLMSettings",
    "Memory",
    "Messages",
    "Packet",
    "Packet",
    "ParallelProcessor",
    "Printer",
    "ProcName",
    "Processor",
    "RunContext",
    "SystemMessage",
    "UserMessage",
    "print_event_stream",
]



import logging
from collections.abc import AsyncIterator, Sequence
from functools import partial
from typing import Any, Generic, Literal
from uuid import uuid4

from grasp_agents.tracing_decorators import workflow

from .errors import RunnerError
from .event_bus import EventBus
from .packet import Packet
from .processors.base_processor import BaseProcessor
from .run_context import CtxT, RunContext
from .typing.events import Event, ProcPacketOutEvent, RunPacketOutEvent
from .typing.io import OutT

logger = logging.getLogger(__name__)

START_PROC_NAME: Literal["*START*"] = "*START*"
END_PROC_NAME: Literal["*END*"] = "*END*"


class Runner(Generic[OutT, CtxT]):
    def __init__(
        self,
        entry_proc: BaseProcessor[Any, Any, CtxT],
        procs: Sequence[BaseProcessor[Any, Any, CtxT]],
        ctx: RunContext[CtxT] | None = None,
        name: str | None = None,
    ) -> None:
        if entry_proc not in procs:
            raise RunnerError(
                f"Entry processor {entry_proc.name} must be in the list of processors: "
                f"{', '.join(proc.name for proc in procs)}"
            )
        if sum(1 for proc in procs if END_PROC_NAME in (proc.recipients or [])) != 1:
            raise RunnerError(
                "There must be exactly one processor with recipient 'END'."
            )

        self._name = name or str(uuid4())[:6]

        self._entry_proc = entry_proc
        self._procs = procs

        self._event_bus = EventBus()

        self._ctx = ctx or RunContext[CtxT](state=None)  # type: ignore

    @property
    def name(self) -> str:
        return self._name

    @property
    def ctx(self) -> RunContext[CtxT]:
        return self._ctx

    def _generate_call_id(self, proc: BaseProcessor[Any, Any, CtxT]) -> str | None:
        return self._name + "/" + proc.generate_call_id(call_id=None)

    def _make_start_event(self, chat_inputs: Any) -> ProcPacketOutEvent:
        start_packet = Packet[Any](
            sender=START_PROC_NAME,
            routing=[[self._entry_proc.name]],
            payloads=[chat_inputs],
        )
        return ProcPacketOutEvent(
            id=start_packet.id,
            data=start_packet,
            src_name=START_PROC_NAME,
            dst_name=self._entry_proc.name,
            call_id=None,
        )

    def _unpack_packet(
        self, packet: Packet[Any]
    ) -> tuple[Packet[Any] | None, Any | None]:
        if packet.sender == START_PROC_NAME:
            return None, packet.payloads[0]
        return packet, None

    async def _event_handler(
        self,
        in_event: Event[Any],
        *,
        proc: BaseProcessor[Any, Any, CtxT],
        **run_kwargs: Any,
    ) -> None:
        if not (isinstance(in_event, ProcPacketOutEvent)):
            # Currently, we only handle ProcResultEvent in the runner
            # TODO: User handoffs and MCP tool calls
            return

        logger.info(f"\n[Running processor {proc.name}]\n")

        in_packet, chat_inputs = self._unpack_packet(in_event.data)
        call_id = self._generate_call_id(proc)
        out_packet: Packet[Any] | None = None

        async for out_event in proc.run_stream(
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            ctx=self._ctx,
            call_id=call_id,
            **run_kwargs,
        ):
            if (
                isinstance(out_event, ProcPacketOutEvent)
                and out_event.src_name == proc.name
            ):
                out_packet = out_event.data

                if out_packet.routing == [[END_PROC_NAME]]:
                    final_event = RunPacketOutEvent(
                        id=out_packet.id,
                        data=out_packet,
                        src_name=out_packet.sender,
                        dst_name=END_PROC_NAME,
                        call_id=call_id,
                    )
                    await self._event_bus.push_to_stream(final_event)
                    await self._event_bus.finalize(final_event.data)
                    break

                for sub_out_packet in out_packet.split_by_recipient() or []:
                    if not sub_out_packet.routing or not sub_out_packet.routing[0]:
                        continue

                    dst_name = sub_out_packet.routing[0][0]
                    sub_out_event = ProcPacketOutEvent(
                        id=sub_out_packet.id,
                        data=sub_out_packet,
                        src_name=sub_out_packet.sender,
                        dst_name=dst_name,
                        call_id=call_id,
                    )
                    await self._event_bus.push_to_stream(sub_out_event)
                    await self._event_bus.post(sub_out_event)

            else:
                await self._event_bus.push_to_stream(out_event)

        if out_packet is None:
            # Cannot happen, for type checking purposes
            return

        route = out_packet.uniform_routing or out_packet.routing
        logger.info(
            f"\n[Finished running processor {proc.name}]\n"
            f"Posted output packet to recipients: {route}\n"
        )

    @workflow(name="runner_run")  # type: ignore
    async def run_stream(
        self, chat_inputs: Any = "start", **run_kwargs: Any
    ) -> AsyncIterator[Event[Any]]:
        async with self._event_bus:
            for proc in self._procs:
                self._event_bus.register_event_handler(
                    dst_name=proc.name,
                    handler=partial(self._event_handler, proc=proc, **run_kwargs),
                )

            await self._event_bus.post(self._make_start_event(chat_inputs))

            async for event in self._event_bus.stream_events():
                yield event

    async def run(self, chat_inputs: Any = "start", **run_kwargs: Any) -> Packet[OutT]:
        async for _ in self.run_stream(chat_inputs=chat_inputs, **run_kwargs):
            pass
        return await self._event_bus.final_result()

    async def shutdown(self) -> None:
        await self._event_bus.shutdown()



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Sequence
from typing import Any

from ..errors import WorkflowConstructionError
from ..processors.processor import Processor
from ..run_context import CtxT, RunContext
from ..typing.events import DummyEvent, Event
from ..typing.io import InT, OutT, ProcName


class WorkflowProcessor(Processor[InT, OutT, CtxT], ABC):
    def __init__(
        self,
        name: ProcName,
        subprocs: Sequence[Processor[Any, Any, CtxT]],
        start_proc: Processor[InT, Any, CtxT],
        end_proc: Processor[Any, OutT, CtxT],
        recipients: Sequence[ProcName] | None = None,
    ) -> None:
        super().__init__(
            name=name, recipients=(recipients or end_proc.recipients), max_retries=0
        )

        self._in_type = start_proc.in_type
        self._out_type = end_proc.out_type

        if len(subprocs) < 2:
            raise WorkflowConstructionError("At least two subprocessors are required")
        if start_proc not in subprocs:
            raise WorkflowConstructionError(
                "Start subprocessor must be in the subprocessors list"
            )
        if end_proc not in subprocs:
            raise WorkflowConstructionError(
                "End subprocessor must be in the subprocessors list"
            )

        self._subprocs = subprocs
        self._start_proc = start_proc
        self._end_proc = end_proc

        for subproc in subprocs:
            subproc.recipients = None

    def select_recipients_impl(
        self, output: OutT, *, ctx: RunContext[CtxT], call_id: str
    ) -> Sequence[ProcName]:
        return self._end_proc.select_recipients_impl(
            output=output, ctx=ctx, call_id=call_id
        )

    @property
    def subprocs(self) -> Sequence[Processor[Any, Any, CtxT]]:
        return self._subprocs

    @property
    def start_proc(self) -> Processor[InT, Any, CtxT]:
        return self._start_proc

    @property
    def end_proc(self) -> Processor[Any, OutT, CtxT]:
        return self._end_proc

    @abstractmethod
    async def _process(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> list[OutT]:
        pass

    @abstractmethod
    async def _process_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> AsyncIterator[Event[Any]]:
        yield DummyEvent()



import contextlib
import logging
from collections.abc import AsyncGenerator, AsyncIterator, Sequence
from typing import Any, ClassVar, Generic, TypeVar, cast, final

from pydantic import TypeAdapter
from pydantic import ValidationError as PydanticValidationError
from typing_extensions import Protocol

from grasp_agents.errors import (
    PacketRoutingError,
    ProcInputValidationError,
    ProcOutputValidationError,
)
from grasp_agents.tracing_decorators import workflow

from ..packet import Packet
from ..run_context import CtxT, RunContext
from ..typing.events import Event, ProcPacketOutEvent, ProcPayloadOutEvent
from ..typing.io import InT, OutT, ProcName
from ..utils.callbacks import is_method_overridden
from ..utils.streaming import wrap_with_aclosing
from .base_processor import BaseProcessor, with_retry

logger = logging.getLogger(__name__)


_OutT_contra = TypeVar("_OutT_contra", contravariant=True)


class RecipientSelector(Protocol[_OutT_contra, CtxT]):
    def __call__(
        self, output: _OutT_contra, *, ctx: RunContext[CtxT], call_id: str
    ) -> Sequence[ProcName]: ...


class Processor(BaseProcessor[InT, OutT, CtxT], Generic[InT, OutT, CtxT]):
    """
    Processor that can have different numbers of inputs and outputs, allowing for an
    arbitrary mapping between them.
    """

    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def validate_inputs(
        self,
        call_id: str,
        chat_inputs: Any | None = None,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
    ) -> list[InT] | None:
        err_kwargs = {"proc_name": self.name, "call_id": call_id}

        num_non_null_inputs = sum(
            x is not None for x in [chat_inputs, in_args, in_packet]
        )

        if num_non_null_inputs > 1:
            raise ProcInputValidationError(
                message="Only one of chat_inputs, in_args, or in_message must be provided",
                **err_kwargs,
            )
        if self.in_type is not type(None) and num_non_null_inputs == 0:
            raise ProcInputValidationError(
                message="One of chat_inputs, in_args, or in_message must be provided",
                **err_kwargs,
            )

        if in_packet is not None and not in_packet.payloads:
            raise ProcInputValidationError(
                message="in_packet must contain at least one payload", **err_kwargs
            )
        if in_args is not None and not in_args:
            raise ProcInputValidationError(
                message="in_args must contain at least one argument", **err_kwargs
            )

        if chat_inputs is not None:
            # 1) chat_inputs are provided -> no need to validate further
            return None

        resolved_args: list[InT]

        if isinstance(in_args, self.in_type):
            # 2) Single in_args of correct type is provided
            resolved_args = [in_args]

        elif isinstance(in_args, list):
            # 3) List of in_args is provided
            resolved_args = cast("list[InT]", in_args)

        elif in_args is not None:
            raise ProcInputValidationError(
                message=f"in_args are neither of type {self.in_type} "
                f"nor a list of {self.in_type}.",
                **err_kwargs,
            )

        else:
            # 4) in_packet is provided
            resolved_args = list(cast("Packet[InT]", in_packet).payloads)

        try:
            for args in resolved_args:
                TypeAdapter(self._in_type).validate_python(args)
        except PydanticValidationError as err:
            raise ProcInputValidationError(message=str(err), **err_kwargs) from err

        return resolved_args

    def validate_output(self, out_payload: OutT, call_id: str) -> OutT:
        if out_payload is None:
            return out_payload
        try:
            return TypeAdapter(self.out_type).validate_python(out_payload)
        except PydanticValidationError as err:
            raise ProcOutputValidationError(
                schema=self.out_type, proc_name=self.name, call_id=call_id
            ) from err

    def _validate_recipients(
        self, recipients: Sequence[ProcName] | None, call_id: str
    ) -> None:
        for r in recipients or []:
            if r not in (self.recipients or []):
                raise PacketRoutingError(
                    proc_name=self.name,
                    call_id=call_id,
                    selected_recipient=r,
                    allowed_recipients=cast("list[str]", self.recipients),
                )

    def select_recipients_impl(
        self, output: OutT, *, ctx: RunContext[CtxT], call_id: str
    ) -> Sequence[ProcName]:
        raise NotImplementedError

    def add_recipient_selector(
        self, func: RecipientSelector[OutT, CtxT]
    ) -> RecipientSelector[OutT, CtxT]:
        self.select_recipients_impl = func

        return func

    @final
    def select_recipients(
        self, output: OutT, ctx: RunContext[CtxT], call_id: str
    ) -> Sequence[ProcName]:
        base_cls = BaseProcessor[Any, Any, Any]
        if is_method_overridden("select_recipients_impl", self, base_cls):
            recipients = self.select_recipients_impl(
                output=output, ctx=ctx, call_id=call_id
            )
            self._validate_recipients(recipients, call_id=call_id)
            return recipients

        # self.processor is not None because otherwise we do not call this method
        return cast("list[ProcName]", self.recipients)

    async def _process(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> list[OutT]:
        """
        Process a list of inputs and return a list of outputs. The length of the
        output list can be different from the input list.
        """
        return cast("list[OutT]", in_args)

    async def _process_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> AsyncIterator[Event[Any]]:
        outputs = await self._process(
            chat_inputs=chat_inputs, in_args=in_args, call_id=call_id, ctx=ctx
        )
        for output in outputs:
            yield ProcPayloadOutEvent(data=output, src_name=self.name, call_id=call_id)

    def _build_packet(
        self, outputs: list[OutT], call_id: str, ctx: RunContext[CtxT]
    ) -> Packet[OutT]:
        for output in outputs:
            self.validate_output(output, call_id=call_id)

        routings: list[Sequence[ProcName]] | None = []
        if self.recipients is not None:
            for output in outputs:
                routings.append(
                    self.select_recipients(output=output, ctx=ctx, call_id=call_id)
                )

        joined_routing = [r for r in routings] if routings else None

        return Packet(sender=self.name, payloads=outputs, routing=joined_routing)

    @workflow(name="processor")  # type: ignore
    @final
    @with_retry
    @wrap_with_aclosing
    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncGenerator[Event[Any], None]:
        ctx = ctx or RunContext[CtxT](state=None)  # type: ignore
        call_id = self.generate_call_id(call_id)

        val_in_args = self.validate_inputs(
            call_id=call_id,
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
        )

        outputs: list[OutT] = []
        async for event in self._process_stream(
            chat_inputs=chat_inputs,
            in_args=val_in_args,
            call_id=call_id,
            ctx=ctx,
        ):
            if isinstance(event, ProcPayloadOutEvent) and event.src_name == self.name:
                # Collect payloads of the processor's own output events
                # This makes sure subprocessor events are yielded but their
                # payloads are not included in the final output packet
                outputs.append(event.data)
            else:
                yield event

        # 1) Combine payloads
        # 2) Obtain recipients per payload
        # 3) Form a packet
        # 4) Yield a ProcPacketOutEvent
        out_packet = self._build_packet(outputs=outputs, call_id=call_id, ctx=ctx)
        yield ProcPacketOutEvent(
            id=out_packet.id, data=out_packet, src_name=self.name, call_id=call_id
        )

    @final
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT]:
        result = None
        async for event in self.run_stream(
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
            call_id=call_id,
            ctx=ctx,
        ):
            if isinstance(event, ProcPacketOutEvent) and event.src_name == self.name:
                result = event.data
                break
        if result is None:
            raise RuntimeError("Processor run did not yield a ProcPacketOutputEvent")
        return result



from logging import getLogger
from pathlib import Path

logger = getLogger(__name__)


def read_txt(file_path: str | Path, encoding: str = "utf-8") -> str:
    return Path(file_path).read_text(encoding=encoding)


def read_contents_from_file(
    file_path: str | Path,
    binary_mode: bool = False,
) -> str | bytes:
    try:
        if binary_mode:
            return Path(file_path).read_bytes()
        return Path(file_path).read_text()
    except FileNotFoundError:
        logger.exception(f"File {file_path} not found.")
        return ""


def get_prompt(prompt_text: str | None, prompt_path: str | Path | None) -> str | None:
    if prompt_text is None:
        return read_contents_from_file(prompt_path) if prompt_path is not None else None  # type: ignore[arg-type]

    return prompt_text
